"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.Address = exports.LocalAddress = void 0;
const tslib_1 = require("tslib");
const ripemd160_1 = tslib_1.__importDefault(require("ripemd160"));
const ethereumjs_util_1 = require("ethereumjs-util");
const crypto_utils_1 = require("./crypto-utils");
const pb = tslib_1.__importStar(require("./proto/loom_pb"));
class LocalAddress {
    constructor(bytes) {
        this.bytes = bytes;
        this.bytes = Buffer.from(bytes);
    }
    isEmpty() {
        return this.bytes && this.bytes.length === 0;
    }
    toString() {
        // TODO: checksum encoding like go-loom
        return ('0x' +
            Buffer.from(this.bytes.buffer, this.bytes.byteOffset, this.bytes.byteLength).toString('hex'));
    }
    toChecksumString() {
        return (0, ethereumjs_util_1.toChecksumAddress)(this.toString());
    }
    equals(other) {
        return Buffer.compare(this.bytes, other.bytes) === 0;
    }
    static fromHexString(hexAddr) {
        if (!hexAddr.startsWith('0x')) {
            throw new Error('hexAddr argument has no 0x prefix');
        }
        const bytes = Buffer.from(hexAddr.slice(2), 'hex');
        if (bytes.length !== 20) {
            throw new Error(`Invalid local address, expected 20 bytes, got ${bytes.length}`);
        }
        return new LocalAddress(bytes);
    }
    /**
     * Converts a public key to a local address (which is used as unique identifier within a DAppChain).
     * @param publicKey 32-byte public key.
     * @returns Array of bytes representing a local address.
     */
    static fromPublicKey(publicKey) {
        if (publicKey.length !== 32) {
            throw new Error(`Invalid public key, expected 32 bytes, go ${publicKey.length}`);
        }
        const hasher = new ripemd160_1.default();
        hasher.update(Buffer.from(publicKey.buffer, publicKey.byteOffset, publicKey.byteLength));
        return new LocalAddress(hasher.digest());
    }
}
exports.LocalAddress = LocalAddress;
class Address {
    constructor(chainId, local) {
        this.chainId = chainId;
        this.local = local;
    }
    isEmpty() {
        return this.chainId === '' && this.local.isEmpty();
    }
    toString() {
        return `${this.chainId}:${this.local.toString()}`;
    }
    MarshalPB() {
        const addr = new pb.Address();
        addr.setChainId(this.chainId);
        addr.setLocal((0, crypto_utils_1.bufferToProtobufBytes)(this.local.bytes));
        return addr;
    }
    equals(other) {
        return this.chainId === other.chainId && this.local.equals(other.local);
    }
    /**
     * @deprecated Use the function UnmarshalPB instead
     */
    static UmarshalPB(pb) {
        console.warn('Function UmarshalPB is deprecated and should be replaced by UnmarshalPB');
        return Address.UnmarshalPB(pb);
    }
    static UnmarshalPB(pb) {
        return new Address(pb.getChainId(), new LocalAddress(pb.getLocal_asU8()));
    }
    /**
     * Converts a string to an address.
     * @param address String representation of an address, in the format "chain:0x...".
     */
    static fromString(address) {
        const parts = address.split(':');
        if (parts.length !== 2) {
            throw new Error('Invalid address string');
        }
        return new Address(parts[0], LocalAddress.fromHexString(parts[1]));
    }
}
exports.Address = Address;
//# sourceMappingURL=address.js.map