"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.Client = exports.TxSyncBroadcaster = exports.TxCommitBroadcaster = exports.isTxAlreadyInCacheError = exports.TX_ALREADY_EXISTS_ERROR = exports.ClientEvent = void 0;
const tslib_1 = require("tslib");
const debug_1 = tslib_1.__importDefault(require("debug"));
const events_1 = tslib_1.__importDefault(require("events"));
const retry_1 = tslib_1.__importDefault(require("retry"));
const loom_pb_1 = require("./proto/loom_pb");
const evm_pb_1 = require("./proto/evm_pb");
const crypto_utils_1 = require("./crypto-utils");
const address_1 = require("./address");
const ws_rpc_client_1 = require("./internal/ws-rpc-client");
const json_rpc_client_1 = require("./internal/json-rpc-client");
const helpers_1 = require("./helpers");
const debugLog = (0, debug_1.default)('client');
var ClientEvent;
(function (ClientEvent) {
    /**
     * Emitted when an event is received from a smart contract.
     * Listener will receive IChainEventArgs.
     */
    ClientEvent["Contract"] = "contractEvent";
    /**
     * Exclusively used by loom-provider
     */
    ClientEvent["EVMEvent"] = "evmEvent";
    /**
     * Emitted when an error occurs that can't be relayed by other means.
     * Listener will receive IClientErrorEventArgs.
     */
    ClientEvent["Error"] = "error";
    /**
     * Emitted when a connection is established to the DAppChain.
     * Listener will receive INetEventArgs.
     */
    ClientEvent["Connected"] = "connected";
    /**
     * Emitted when a connection with the DAppChain is closed.
     * Listener will receive INetEventArgs.
     */
    ClientEvent["Disconnected"] = "disconnected";
})(ClientEvent = exports.ClientEvent || (exports.ClientEvent = {}));
exports.TX_ALREADY_EXISTS_ERROR = 'Tx already exists in cache';
function isTxAlreadyInCacheError(err) {
    // TODO: Need to update the WS client to throw the same errors as the HTTP client, so don't
    //       have to detect two different errors everywhere.
    return ((err instanceof Error && err.message.indexOf(exports.TX_ALREADY_EXISTS_ERROR) !== -1) || // HTTP
        (err.data && err.data.indexOf(exports.TX_ALREADY_EXISTS_ERROR) !== -1) // WS
    );
}
exports.isTxAlreadyInCacheError = isTxAlreadyInCacheError;
class TxCommitBroadcaster {
    broadcast(client, txBytes) {
        return client.sendAsync('broadcast_tx_commit', [
            (0, crypto_utils_1.Uint8ArrayToB64)(txBytes)
        ]);
    }
}
exports.TxCommitBroadcaster = TxCommitBroadcaster;
class TxSyncBroadcaster {
    constructor() {
        this.resultPollingStrategy = {
            retries: 5,
            minTimeout: 3000,
            maxTimeout: 5000,
            randomize: true
        };
    }
    broadcast(client, txBytes) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const checkTxResult = yield client.sendAsync('broadcast_tx_sync', [
                (0, crypto_utils_1.Uint8ArrayToB64)(txBytes)
            ]);
            const { code, log, data } = checkTxResult;
            // if the tx failed in CheckTx it won't make it into a block, and won't be indexed
            if (code !== 0) {
                return {
                    check_tx: { code, log, data },
                    deliver_tx: {},
                    hash: checkTxResult.hash
                };
            }
            yield (0, helpers_1.sleep)(this.resultPollingStrategy.minTimeout);
            const op = retry_1.default.operation(this.resultPollingStrategy);
            const result = yield new Promise((resolve, reject) => {
                op.attempt((currentAttempt) => {
                    debugLog(`Querying for result of tx ${checkTxResult.hash} - attempt ${currentAttempt}`);
                    client
                        .sendAsync('tx', {
                        hash: Buffer.from(checkTxResult.hash, 'hex').toString('base64')
                    })
                        .then(result => resolve(result))
                        .catch(err => {
                        debugLog(`Failed to retrieve result of tx ${checkTxResult.hash}: ${err.message || err.data}`);
                        // keep trying to retrieve the result if the tx isn't found, until all retries are used up
                        if ((err instanceof Error && err.message.endsWith('not found')) || // HTTP
                            (err.data && err.data.endsWith('not found')) // WS
                        ) {
                            if (!op.retry(err)) {
                                reject(err);
                            }
                        }
                        else {
                            reject(err);
                        }
                    });
                });
            });
            return {
                check_tx: { code, log, data },
                deliver_tx: result.tx_result,
                hash: result.hash,
                height: result.height
            };
        });
    }
}
exports.TxSyncBroadcaster = TxSyncBroadcaster;
/**
 * Writes to & reads from a Loom DAppChain.
 *
 * The client can listen to events emitted by smart contracts running on a DAppChain,
 * there is currently only one type of event. The event subscription API matches the NodeJS
 * EventEmitter API. For example...
 *
 * function subscribeToEvents(client: Client) {
 *   client.on(ClientEvent.Contract, (event: IChainEventArgs) => {
 *     // handle event
 *   }
 * }
 */
class Client extends events_1.default {
    get readUrl() {
        return this._readClient.url;
    }
    get writeUrl() {
        return this._writeClient.url;
    }
    constructor(chainId, writeClient, readClient) {
        super();
        /** Middleware to apply to transactions before they are transmitted to the DAppChain. */
        this.txMiddleware = [];
        this.chainId = chainId;
        this.txBroadcaster = new TxSyncBroadcaster();
        // TODO: basic validation of the URIs to ensure they have all required components.
        this._writeClient =
            typeof writeClient === 'string' ? new ws_rpc_client_1.WSRPCClient(writeClient) : writeClient;
        this._writeClient.on(json_rpc_client_1.RPCClientEvent.Error, (url, err) => this._emitNetEvent(url, ClientEvent.Error, err));
        this._writeClient.on(json_rpc_client_1.RPCClientEvent.Connected, (url) => this._emitNetEvent(url, ClientEvent.Connected));
        this._writeClient.on(json_rpc_client_1.RPCClientEvent.Disconnected, (url) => this._emitNetEvent(url, ClientEvent.Disconnected));
        if (!readClient || writeClient === readClient) {
            this._readClient = this._writeClient;
        }
        else {
            this._readClient = typeof readClient === 'string' ? new ws_rpc_client_1.WSRPCClient(readClient) : readClient;
            this._readClient.on(json_rpc_client_1.RPCClientEvent.Error, (url, err) => this._emitNetEvent(url, ClientEvent.Error, err));
            this._readClient.on(json_rpc_client_1.RPCClientEvent.Connected, (url) => this._emitNetEvent(url, ClientEvent.Connected));
            this._readClient.on(json_rpc_client_1.RPCClientEvent.Disconnected, (url) => this._emitNetEvent(url, ClientEvent.Disconnected));
        }
        this.isWeb3EndpointEnabled = true === /eth$/.test(this._readClient.url);
        const emitContractEvent = (url, event) => this._emitContractEvent(url, event);
        this.on('newListener', (event) => {
            if (event === ClientEvent.Contract && this.listenerCount(ClientEvent.Contract) === 0) {
                this._readClient.on(json_rpc_client_1.RPCClientEvent.Message, emitContractEvent);
            }
        });
        this.on('removeListener', (event) => {
            if (event === ClientEvent.Contract && this.listenerCount(ClientEvent.Contract) === 0) {
                this._readClient.removeListener(json_rpc_client_1.RPCClientEvent.Message, emitContractEvent);
            }
        });
    }
    /**
     * Cleans up all underlying network resources.
     * Once disconnected the client can no longer be used to interact with the DAppChain.
     */
    disconnect() {
        this._writeClient.disconnect();
        if (this._readClient && this._readClient != this._writeClient) {
            this._readClient.disconnect();
        }
    }
    /**
     * Commits a transaction to the DAppChain.
     *
     * Consider using Contract.callAsync() instead.
     *
     * @param tx Transaction to commit.
     * @param opts Options object.
     * @param opts.middleware Middleware to apply before sending the tx to the DAppChain, setting this
     *                        option will override the default set of middleware specified in
     *                        the `Client.txMiddleware` property.
     * @returns Result (if any) returned by the tx handler in the contract that processed the tx.
     */
    commitTxAsync(tx, opts = {}) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const { middleware = this.txMiddleware } = opts;
            let txBytes = tx.serializeBinary();
            for (let i = 0; i < middleware.length; i++) {
                txBytes = yield middleware[i].Handle(txBytes);
            }
            let result;
            try {
                result = yield this.txBroadcaster.broadcast(this._writeClient, txBytes);
            }
            catch (err) {
                // Allow the middleware to handle errors. They're applied in reverse order so that the last
                // middleware that was applied to the tx that was sent will be get to handle the error first.
                for (let i = middleware.length - 1; i >= 0; i--) {
                    if (middleware[i].handleError) {
                        middleware[i].handleError(err);
                    }
                }
                if (isTxAlreadyInCacheError(err)) {
                    throw new Error(exports.TX_ALREADY_EXISTS_ERROR);
                }
                throw err;
            }
            debugLog(`Result ${JSON.stringify(result, null, 2)}`);
            if (result) {
                let results = { validation: result.check_tx, commit: result.deliver_tx };
                // Allow the middleware to detect specific error conditions, and throw more precise errors
                for (let i = middleware.length - 1; i >= 0; i--) {
                    if (middleware[i].HandleResults) {
                        results = middleware[i].HandleResults(results);
                    }
                }
                // Throw generic errors if the tx failed
                if ((result.check_tx.code || 0) != 0) {
                    if (!result.check_tx.log) {
                        throw new Error(`Failed to commit Tx: ${result.check_tx.code}`);
                    }
                    throw new Error(`Failed to commit Tx: ${result.check_tx.log}`);
                }
                if ((result.deliver_tx.code || 0) != 0) {
                    if (!result.deliver_tx.log) {
                        throw new Error(`Failed to commit Tx: ${result.deliver_tx.code}`);
                    }
                    throw new Error(`Failed to commit Tx: ${result.deliver_tx.log}`);
                }
                if (result.deliver_tx.data) {
                    return (0, crypto_utils_1.B64ToUint8Array)(result.deliver_tx.data);
                }
            }
        });
    }
    /**
     * addListenerForTopics
     */
    addListenerForTopics() {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const emitContractEvent = (url, event) => {
                this._emitContractEvent(url, event, true);
            };
            this._readClient.on(json_rpc_client_1.RPCClientEvent.EVMMessage, emitContractEvent);
        });
    }
    /**
     * Queries the current state of a contract.
     *
     * Consider using Contract.staticCallAsync() instead.
     */
    queryAsync(contract, query, vmType = loom_pb_1.VMType.PLUGIN, caller) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const result = yield this._readClient.sendAsync('query', {
                contract: contract.local.toString(),
                query: query ? (0, crypto_utils_1.Uint8ArrayToB64)(query) : undefined,
                vmType: vmType,
                caller: caller ? caller.toString() : undefined
            });
            if (result) {
                return (0, crypto_utils_1.B64ToUint8Array)(result);
            }
        });
    }
    /**
     * Sends a Web3 JSON-RPC message to the /eth endpoint on a Loom node.
     * @param method RPC method name.
     * @param params Parameter object or array.
     * @returns A promise that will be resolved with the value of the result field (if any) in the
     *          JSON-RPC response message.
     */
    sendWeb3MsgAsync(method, params) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            return this._readClient.sendAsync(method, params);
        });
    }
    /**
     * Queries the receipt corresponding to a transaction hash
     *
     * @param txHash Transaction hash returned by call transaction.
     * @return EvmTxReceipt The corresponding transaction receipt.
     */
    getEvmTxReceiptAsync(txHash) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const result = yield this._readClient.sendAsync('evmtxreceipt', {
                txHash: (0, crypto_utils_1.Uint8ArrayToB64)(txHash)
            });
            if (result) {
                return evm_pb_1.EvmTxReceipt.deserializeBinary((0, crypto_utils_1.bufferToProtobufBytes)((0, crypto_utils_1.B64ToUint8Array)(result)));
            }
            else {
                return null;
            }
        });
    }
    /**
     * Returns the information about a transaction requested by transaction hash
     *
     * @param txHash Transaction hash returned by call transaction.
     * @return EvmTxObject The corresponding transaction object data.
     */
    getEvmTxByHashAsync(txHash) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const result = yield this._readClient.sendAsync('getevmtransactionbyhash', {
                txHash: (0, crypto_utils_1.Uint8ArrayToB64)(txHash)
            });
            if (result) {
                return evm_pb_1.EvmTxObject.deserializeBinary((0, crypto_utils_1.bufferToProtobufBytes)((0, crypto_utils_1.B64ToUint8Array)(result)));
            }
            else {
                return null;
            }
        });
    }
    /**
     * Queries the code corresponding to a contract
     *
     * @param contractAddress Contract address returned by deploy.
     * @return Uint8Array The corresponding contract code
     */
    getEvmCodeAsync(contractAddress) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const result = yield this._readClient.sendAsync('getevmcode', {
                contract: contractAddress.toString()
            });
            if (result) {
                return (0, crypto_utils_1.B64ToUint8Array)(result);
            }
            else {
                return null;
            }
        });
    }
    /**
     * Queries logs with filter terms
     *
     * @param filter Filter terms
     * @return Uint8Array The corresponding result of the filter
     */
    getEvmLogsAsync(filterObject) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const filter = JSON.stringify(filterObject);
            debugLog(`Send filter ${filter} to getlogs`);
            const result = yield this._readClient.sendAsync('getevmlogs', {
                filter
            });
            if (result) {
                return (0, crypto_utils_1.B64ToUint8Array)(result);
            }
            else {
                return null;
            }
        });
    }
    /**
     * Creates a new filter based on filter terms, to notify when the state changes
     *
     * The function getEVMNewFilterAsync works in the similar way of the RPC call eth_newFilter, for more
     *
     * Also for understand how filters works check https://github.com/ethereum/wiki/wiki/JSON-RPC#eth_newfilter
     *
     * @param filter Filter terms
     * @return Uint8Array The corresponding result of the filter
     */
    newEvmFilterAsync(filterObject) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const filter = JSON.stringify(filterObject);
            debugLog(`Send filter ${filter} to newfilter`);
            const result = yield this._readClient.sendAsync('newevmfilter', {
                filter
            });
            if (result) {
                return result;
            }
            else {
                return null;
            }
        });
    }
    /**
     * Polling method for a filter, which returns an array of logs which occurred since last poll
     *
     * The ID used was requested from getEVMNewFilterChanges or getEVMNewBlockFilter
     *
     * @param id Id of filter previously created
     * @return Uint8Array The corresponding result of the request for given id
     */
    getEvmFilterChangesAsync(id) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            debugLog(`Get filter changes for ${JSON.stringify({ id }, null, 2)}`);
            const result = yield this._readClient.sendAsync('getevmfilterchanges', {
                id
            });
            if (result) {
                const envelope = evm_pb_1.EthFilterEnvelope.deserializeBinary((0, crypto_utils_1.bufferToProtobufBytes)(result));
                switch (envelope.getMessageCase()) {
                    case evm_pb_1.EthFilterEnvelope.MessageCase.ETH_BLOCK_HASH_LIST:
                        return envelope.getEthBlockHashList();
                    case evm_pb_1.EthFilterEnvelope.MessageCase.ETH_FILTER_LOG_LIST:
                        return envelope.getEthFilterLogList();
                    case evm_pb_1.EthFilterEnvelope.MessageCase.ETH_TX_HASH_LIST:
                        return envelope.getEthTxHashList();
                    case evm_pb_1.EthFilterEnvelope.MessageCase.MESSAGE_NOT_SET:
                    default:
                        return null;
                }
            }
            else {
                return null;
            }
        });
    }
    /**
     * Creates a filter in the node, to notify when a new block arrives
     *
     * In order to check if the state has changed, call getEVMFilterChangesAsync
     *
     * @return String Filter ID in hex format to be used later with getEVMFilterChangesAsync
     */
    newBlockEvmFilterAsync() {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const result = yield this._readClient.sendAsync('newblockevmfilter', {});
            if (result) {
                return result.toString();
            }
            else {
                return null;
            }
        });
    }
    /**
     * Creates a filter in the node, to notify when new pending transactions arrive.
     *
     * In order to check if the state has changed, call getEVMFilterChangesAsync
     *
     * @return String Filter ID in hex format to be used later with getEVMFilterChangesAsync
     */
    newPendingTransactionEvmFilterAsync() {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const result = yield this._readClient.sendAsync('newpendingtransactionevmfilter', {});
            if (result) {
                return result.toString();
            }
            else {
                return null;
            }
        });
    }
    /**
     * Uninstall/delete previously created filters
     *
     * The ID used was requested from getEVMNewFilterChanges or getEVMNewBlockFilter
     *
     * @param id Id of filter previously created
     * @return boolean If true the filter is removed with success
     */
    uninstallEvmFilterAsync(id) {
        return this._readClient.sendAsync('uninstallevmfilter', {
            id
        });
    }
    /**
     * Returns information about a block by block number.
     *
     * @param num Integer of a block number
     * @param full If true it returns the full transaction objects, if false only the hashes of the transactions
     */
    getEvmBlockByNumberAsync(num, full = true) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const result = yield this._readClient.sendAsync('getevmblockbynumber', {
                number: num,
                full
            });
            if (result) {
                return evm_pb_1.EthBlockInfo.deserializeBinary((0, crypto_utils_1.bufferToProtobufBytes)((0, crypto_utils_1.B64ToUint8Array)(result)));
            }
            else {
                return null;
            }
        });
    }
    /**
     * Returns the information about a transaction requested by transaction hash.
     *
     * @param hash String with the hash of the transaction
     * @param full If true it returns the full transaction objects, if false only the hashes of the transactions
     */
    getEvmBlockByHashAsync(hashHexStr, full = true) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const result = yield this._readClient.sendAsync('getevmblockbyhash', {
                hash: Buffer.from(hashHexStr.slice(2), 'hex').toString('base64'),
                full
            });
            if (result) {
                return evm_pb_1.EthBlockInfo.deserializeBinary((0, crypto_utils_1.bufferToProtobufBytes)((0, crypto_utils_1.B64ToUint8Array)(result)));
            }
            else {
                return null;
            }
        });
    }
    /**
     * It works by subscribing to particular events. The node will return a subscription id.
     * For each event that matches the subscription a notification with relevant data is send
     * together with the subscription id.
     *
     * Possible methods:
     *  * "NewHeads": Fires a notification each time a new header is appended to the chain
     *  * "Logs": Returns logs that are included in new imported blocks and match the given filter criteria
     *
     * Example of a "filter" (JSON String) with method "logs":
     *  {
     *    "address": "0xa520fe7702b96808f7bbc0d4a233ed1468216cfd",
     *    "topics": ["0x238a0cb8bb633d06981248b822e7bd33c2a35a6089241d099fa519e361cab902"]
     *  }
     *
     * @param method Method selected to the filter, can be "newHeads" or "logs"
     * @param filter JSON string of the filter
     */
    evmSubscribeAsync(method, params) {
        if (this.isWeb3EndpointEnabled) {
            return this._readClient.sendAsync('eth_subscribe', params);
        }
        const filter = JSON.stringify(params);
        return this._readClient.sendAsync('evmsubscribe', {
            method,
            filter
        });
    }
    /**
     * Subscriptions are cancelled method and the subscription id as first parameter.
     * It returns a bool indicating if the subscription was cancelled successful.
     *
     * @param id Id of subscription previously created
     * @return boolean If true the subscription is removed with success
     */
    evmUnsubscribeAsync(id) {
        if (this.isWeb3EndpointEnabled) {
            return this._readClient.sendAsync('eth_unsubscribe', [id]);
        }
        return this._readClient.sendAsync('evmunsubscribe', { id });
    }
    /**
     * Gets the number of the latest block
     *
     * @return The block height
     */
    getBlockHeightAsync() {
        return this._readClient.sendAsync('getblockheight', {});
    }
    /**
     * Gets a nonce for the account identified by the given public key.
     *
     * This should only be called by middleware.
     *
     * @param key A hex encoded public key.
     * @return The nonce.
     */
    getNonceAsync(key) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            return this.getAccountNonceAsync({ key });
        });
    }
    /**
     * Gets a nonce for the account identified by the given public key or address.
     *
     * Only the key or the account needs to be provided, if both are provided the key is ignored.
     * This should only be called by middleware.
     *
     * @param key A hex encoded public key.
     * @parma account Account address prefixed by the chain ID, in the form chainID:0xdeadbeef
     * @return The nonce.
     */
    getAccountNonceAsync(params) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            return parseInt(yield this._writeClient.sendAsync('nonce', params), 10);
        });
    }
    /**
     * Tries to resolve a contract name to an address.
     *
     * @param contractName Name of a smart contract on a Loom DAppChain.
     * @returns Contract address, or null if a contract matching the given name wasn't found.
     */
    getContractAddressAsync(contractName) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const addrStr = yield this._readClient.sendAsync('resolve', { name: contractName });
            if (!addrStr) {
                return null;
            }
            debugLog(`Found contract ${contractName} with address ${address_1.Address.fromString(addrStr)}`);
            return address_1.Address.fromString(addrStr);
        });
    }
    _emitContractEvent(url, event, isEVM = false) {
        debugLog('_emitContractEvent', arguments);
        const { error, result } = event;
        if (error) {
            const eventArgs = { kind: ClientEvent.Error, url, error };
            this.emit(ClientEvent.Error, eventArgs);
        }
        else if (this.isWeb3EndpointEnabled && isEVM) {
            this.emit(ClientEvent.EVMEvent, event);
        }
        else if (result) {
            // Ugh, no built-in JSON->Protobuf marshaller apparently
            // https://github.com/google/protobuf/issues/1591 so gotta do this manually
            const eventArgs = {
                id: event.id,
                kind: isEVM ? ClientEvent.EVMEvent : ClientEvent.Contract,
                url,
                contractAddress: new address_1.Address(result.address.chain_id, new address_1.LocalAddress((0, crypto_utils_1.B64ToUint8Array)(result.address.local))),
                callerAddress: new address_1.Address(result.caller.chain_id, new address_1.LocalAddress((0, crypto_utils_1.B64ToUint8Array)(result.caller.local))),
                blockHeight: result.block_height,
                data: (0, crypto_utils_1.B64ToUint8Array)(result.encoded_body || '0x0'),
                topics: result.topics,
                transactionHash: result.tx_hash,
                transactionHashBytes: result.tx_hash ? (0, crypto_utils_1.B64ToUint8Array)(result.tx_hash) : new Uint8Array([])
            };
            if (isEVM) {
                this.emit(ClientEvent.EVMEvent, eventArgs);
            }
            else {
                this.emit(ClientEvent.Contract, eventArgs);
            }
        }
    }
    _emitNetEvent(url, kind, error) {
        if (kind === ClientEvent.Error) {
            const eventArgs = { kind, url, error };
            this.emit(kind, eventArgs);
        }
        else {
            const eventArgs = { kind, url };
            this.emit(kind, eventArgs);
        }
    }
}
exports.Client = Client;
//# sourceMappingURL=client.js.map