"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.Contract = void 0;
const tslib_1 = require("tslib");
const google_protobuf_1 = require("google-protobuf");
const events_1 = tslib_1.__importDefault(require("events"));
const debug_1 = tslib_1.__importDefault(require("debug"));
const client_1 = require("./client");
const loom_pb_1 = require("./proto/loom_pb");
const crypto_utils_1 = require("./crypto-utils");
const debugLog = (0, debug_1.default)('contract');
/**
 * The Contract class streamlines interaction with a contract that was deployed on a Loom DAppChain.
 * Each instance of this class is bound to a specific contract, and provides a simple way of calling
 * into and querying that contract.
 *
 * A contract instance can be used to listen to events emitted by the corresponding smart contract,
 * there is currently only one type of event. The event subscription API matches the NodeJS
 * EventEmitter API. For example...
 *
 * function subscribeToEvents(contract: Contract) {
 *   contract.on(Contract.EVENT, (event: IChainEventArgs) => {
 *     const dataStr = Buffer.from(event.data as Buffer).toString('utf8')
 *     const dataObj = JSON.parse(dataStr)
 *     console.log('Contract Event: ' + dataStr)
 *   })
 * }
 */
class Contract extends events_1.default {
    /**
     * @param params Parameters.
     * @param params.contractAddr Address of a contract on the Loom DAppChain.
     * @param params.contractName Name of the contract.
     * @param params.callerAddr: Address of the caller, generated from the public key of the tx signer,
     *                           e.g. `new Address(client.chainId, LocalAddress.fromPublicKey(pubKey))`
     * @param params.client: Client to use to communicate with the contract.
     */
    constructor(params) {
        super();
        this._client = params.client;
        this.name = params.contractName;
        this.address = params.contractAddr;
        this.caller = params.callerAddr;
        const emitContractEvent = this._emitContractEvent.bind(this);
        this.on('newListener', (event) => {
            if (event === Contract.EVENT && this.listenerCount(event) === 0) {
                this._client.on(client_1.ClientEvent.Contract, emitContractEvent);
            }
        });
        this.on('removeListener', (event) => {
            if (event === Contract.EVENT && this.listenerCount(event) === 0) {
                this._client.removeListener(client_1.ClientEvent.Contract, emitContractEvent);
            }
        });
    }
    /**
     * Calls a contract method that mutates state.
     * The call into the contract is accomplished by committing a tx to the DAppChain.
     * @param method Contract method to call.
     * @param args Arguments to pass to the contract method.
     * @returns A promise that will be resolved with return value (if any) of the contract method.
     */
    callAsync(method, args, output) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const methodTx = new loom_pb_1.ContractMethodCall();
            methodTx.setMethod(method);
            methodTx.setArgs(args.serializeBinary());
            const request = new loom_pb_1.Request();
            request.setContentType(loom_pb_1.EncodingType.PROTOBUF3);
            request.setAccept(loom_pb_1.EncodingType.PROTOBUF3);
            request.setBody(methodTx.serializeBinary());
            const callTx = new loom_pb_1.CallTx();
            callTx.setVmType(loom_pb_1.VMType.PLUGIN);
            callTx.setInput(request.serializeBinary());
            const msgTx = new loom_pb_1.MessageTx();
            msgTx.setFrom(this.caller.MarshalPB());
            msgTx.setTo(this.address.MarshalPB());
            msgTx.setData(callTx.serializeBinary());
            const tx = new loom_pb_1.Transaction();
            tx.setId(2);
            tx.setData(msgTx.serializeBinary());
            debugLog(`call '${method} on ${this.address.toString()} from ${this.caller.toString()}`);
            const result = yield this._client.commitTxAsync(tx);
            if (result && output) {
                const resp = loom_pb_1.Response.deserializeBinary((0, crypto_utils_1.bufferToProtobufBytes)(result));
                const msgClass = output.constructor;
                google_protobuf_1.Message.copyInto(msgClass.deserializeBinary(resp.getBody_asU8()), output);
            }
            return output;
        });
    }
    /**
     * Calls a contract method that doesn't mutate state.
     * This method is usually used to query the current contract state, it doesn't commit any txs.
     * @param method Contract method to call.
     * @param args Arguments to pass to the contract method.
     * @returns A promise that will be resolved with the return value of the contract method.
     */
    staticCallAsync(method, args, output) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const query = new loom_pb_1.ContractMethodCall();
            query.setMethod(method);
            query.setArgs(args.serializeBinary());
            debugLog(`static call '${method} on ${this.address.toString()} from ${this.caller.toString()}`);
            const result = yield this._client.queryAsync(this.address, query.serializeBinary(), loom_pb_1.VMType.PLUGIN, this.caller);
            if (result && output) {
                const msgClass = output.constructor;
                google_protobuf_1.Message.copyInto(msgClass.deserializeBinary((0, crypto_utils_1.bufferToProtobufBytes)(result)), output);
            }
            return output;
        });
    }
    _emitContractEvent(event) {
        if (event.contractAddress.equals(this.address)) {
            this.emit(Contract.EVENT, event);
        }
    }
}
exports.Contract = Contract;
Contract.EVENT = 'event';
//# sourceMappingURL=contract.js.map