"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.BinanceTransferGateway = void 0;
const tslib_1 = require("tslib");
const transfer_gateway_1 = require("./transfer-gateway");
const address_1 = require("../address");
const transfer_gateway_pb_1 = require("../proto/transfer_gateway_pb");
const big_uint_1 = require("../big-uint");
const binanceRootAddress = new address_1.Address('binance', address_1.LocalAddress.fromHexString('0x0000000000000000000000000000000000000000'));
class BinanceTransferGateway extends transfer_gateway_1.TransferGateway {
    static createAsync(client, callerAddr) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const contractAddr = yield client.getContractAddressAsync('binance-gateway');
            if (!contractAddr) {
                throw Error('Failed to resolve contract address for Binance TransferGateway');
            }
            return new BinanceTransferGateway({ contractAddr, callerAddr, client });
        });
    }
    /**
     * Sends a request to the DAppChain Gateway to begin withdrawal of LOOM tokens from the current
     * DAppChain account to an account on Binance Chain. Before calling this method the user
     * attempting to withdraw the tokens must approve the DAppChain Gateway to transfer the given
     * amount plus the withdrawal fee (which is always charged in BNB).
     * @param amount Amount to withdraw (not including the withdrawal fee).
     * @param recipient Binance address of the account the tokens should be withdrawn to.
     * @returns A promise that will be resolved when the DAppChain Gateway has accepted the withdrawal
     *          request.
     */
    withdrawLoomAsync(amount, recipient) {
        const req = new transfer_gateway_pb_1.TransferGatewayWithdrawLoomCoinRequest();
        req.setAmount((0, big_uint_1.marshalBigUIntPB)(amount));
        req.setRecipient(recipient.MarshalPB());
        // TODO: It should be necessary to specify the token contract in the request, but validation
        //       checks in the gateway contract will fail if it's missing, so until the contract is
        //       fixed up we need to specify some kind of address.
        req.setTokenContract(binanceRootAddress.MarshalPB());
        return this.callAsync('WithdrawLoomCoin', req);
    }
    /**
     * Sends a request to the DAppChain Gateway to begin withdrawal of BEP2 tokens from the current
     * DAppChain account to an account on Binance Chain. Before calling this method the user
     * attempting to withdraw the tokens must approve the DAppChain Gateway to transfer the given
     * amount plus the withdrawal fee (which is always charged in BNB).
     * @param amount Amount to withdraw (not including the withdrawal fee).
     * @param tokenContract DAppChain address of BEP2 contract.
     * @param recipient Binance address of the account the tokens should be withdrawn to.
     * @returns A promise that will be resolved when the DAppChain Gateway has accepted the withdrawal
     *          request.
     */
    withdrawTokenAsync(amount, tokenContract, recipient) {
        const req = new transfer_gateway_pb_1.TransferGatewayWithdrawTokenRequest();
        req.setTokenKind(transfer_gateway_pb_1.TransferGatewayTokenKind.BEP2);
        req.setTokenAmount((0, big_uint_1.marshalBigUIntPB)(amount));
        req.setTokenContract(tokenContract.MarshalPB());
        req.setRecipient(recipient.MarshalPB());
        return this.callAsync('WithdrawToken', req);
    }
    /**
     * Sends a request to the DAppChain Gateway to resubmit a previously rejected token withdrawal
     * from the current DAppChain account. This is currently only supported by the Binance gateway.
     * Only the original withdrawer can resubmit a reject withdrawal.
     */
    resubmitWithdrawalAsync() {
        const req = new transfer_gateway_pb_1.TransferGatewayResubmitWithdrawalRequest();
        return this.callAsync('ResubmitWithdrawal', req);
    }
}
exports.BinanceTransferGateway = BinanceTransferGateway;
//# sourceMappingURL=binance-gateway.js.map