"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.DPOS3 = void 0;
const tslib_1 = require("tslib");
const bn_js_1 = tslib_1.__importDefault(require("bn.js"));
const contract_1 = require("../contract");
const address_1 = require("../address");
const dposv3_pb_1 = require("../proto/dposv3_pb");
const big_uint_1 = require("../big-uint");
class DPOS3 extends contract_1.Contract {
    static createAsync(client, callerAddr) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const contractAddr = yield client.getContractAddressAsync('dposV3');
            if (!contractAddr) {
                throw Error('Failed to resolve contract address');
            }
            return new DPOS3({ contractAddr, callerAddr, client });
        });
    }
    constructor(params) {
        super(params);
    }
    getTimeUntilElectionAsync() {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const timeUntilElectionRequest = new dposv3_pb_1.TimeUntilElectionRequestV3();
            const result = yield this.staticCallAsync('TimeUntilElection', timeUntilElectionRequest, new dposv3_pb_1.TimeUntilElectionResponseV3());
            return new bn_js_1.default(result.getTimeUntilElection());
        });
    }
    getCandidatesAsync() {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const listCandidatesReq = new dposv3_pb_1.ListCandidatesRequest();
            const result = yield this.staticCallAsync('ListCandidates', listCandidatesReq, new dposv3_pb_1.ListCandidatesResponse());
            return result.getCandidatesList().map((candidate) => {
                const c = candidate.getCandidate();
                // statistic may not exist yet for a candidate that was just registered (without being whitelisted)
                const s = candidate.getStatistic();
                return {
                    address: address_1.Address.UnmarshalPB(c.getAddress()),
                    pubKey: c.getPubKey_asU8(),
                    delegationTotal: s && s.getDelegationTotal() ? (0, big_uint_1.unmarshalBigUIntPB)(s.getDelegationTotal()) : new bn_js_1.default(0),
                    slashPercentage: s && s.getSlashPercentage() ? (0, big_uint_1.unmarshalBigUIntPB)(s.getSlashPercentage()) : new bn_js_1.default(0),
                    whitelistAmount: s && s.getWhitelistAmount() ? (0, big_uint_1.unmarshalBigUIntPB)(s.getWhitelistAmount()) : new bn_js_1.default(0),
                    whitelistLocktimeTier: s ? s.getLocktimeTier() : dposv3_pb_1.LocktimeTier.TIER_ZERO,
                    maxReferralPercentage: c.getMaxReferralPercentage(),
                    fee: new bn_js_1.default(c.getFee()),
                    newFee: new bn_js_1.default(c.getNewFee()),
                    candidateState: c.getState(),
                    name: c.getName(),
                    description: c.getDescription(),
                    website: c.getWebsite()
                };
            });
        });
    }
    getValidatorsAsync() {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const listValidatorReq = new dposv3_pb_1.ListValidatorsRequest();
            const result = yield this.staticCallAsync('ListValidators', listValidatorReq, new dposv3_pb_1.ListValidatorsResponse());
            return result.getStatisticsList().map((validator) => ({
                address: address_1.Address.UnmarshalPB(validator.getAddress()),
                recentlyMissedBlocks: validator.getRecentlyMissedBlocks(),
                whitelistAmount: validator.getWhitelistAmount()
                    ? (0, big_uint_1.unmarshalBigUIntPB)(validator.getWhitelistAmount())
                    : new bn_js_1.default(0),
                whitelistLocktimeTier: validator.getLocktimeTier(),
                slashPercentage: validator.getSlashPercentage()
                    ? (0, big_uint_1.unmarshalBigUIntPB)(validator.getSlashPercentage())
                    : new bn_js_1.default(0),
                delegationTotal: validator.getDelegationTotal()
                    ? (0, big_uint_1.unmarshalBigUIntPB)(validator.getDelegationTotal())
                    : new bn_js_1.default(0),
                jailed: validator.getJailed()
            }));
        });
    }
    getDelegations(candidate) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const listDelegationsReq = new dposv3_pb_1.ListDelegationsRequestV3();
            listDelegationsReq.setCandidate(candidate.MarshalPB());
            const result = yield this.staticCallAsync('ListDelegations', listDelegationsReq, new dposv3_pb_1.ListDelegationsResponseV3());
            return {
                delegationTotal: result.getDelegationTotal()
                    ? (0, big_uint_1.unmarshalBigUIntPB)(result.getDelegationTotal())
                    : new bn_js_1.default(0),
                delegationsArray: result.getDelegationsList().map(this.getDelegation)
            };
        });
    }
    getAllDelegations() {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const listAllDelegationsReq = new dposv3_pb_1.ListAllDelegationsRequestV3();
            const result = yield this.staticCallAsync('ListAllDelegations', listAllDelegationsReq, new dposv3_pb_1.ListAllDelegationsResponseV3());
            return result.getListResponsesList().map(d => {
                return {
                    delegationTotal: d.getDelegationTotal()
                        ? (0, big_uint_1.unmarshalBigUIntPB)(d.getDelegationTotal())
                        : new bn_js_1.default(0),
                    delegationsArray: d.getDelegationsList().map(this.getDelegation)
                };
            });
        });
    }
    // gets all the delegations by a delegator
    checkAllDelegationsAsync(delegator) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const checkDelegatorDelegationsReq = new dposv3_pb_1.CheckAllDelegationsRequestV3();
            checkDelegatorDelegationsReq.setDelegatorAddress(delegator.MarshalPB());
            const result = yield this.staticCallAsync('CheckAllDelegations', checkDelegatorDelegationsReq, new dposv3_pb_1.CheckAllDelegationsResponseV3());
            return {
                amount: result.getAmount() ? (0, big_uint_1.unmarshalBigUIntPB)(result.getAmount()) : new bn_js_1.default(0),
                weightedAmount: result.getWeightedAmount()
                    ? (0, big_uint_1.unmarshalBigUIntPB)(result.getWeightedAmount())
                    : new bn_js_1.default(0),
                delegationsArray: result.getDelegationsList().map(this.getDelegation)
            };
        });
    }
    // get all the delegations by a delegator to a specific validator
    checkDelegationAsync(validator, delegator) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const checkDelegationReq = new dposv3_pb_1.CheckDelegationRequest();
            checkDelegationReq.setValidatorAddress(validator.MarshalPB());
            checkDelegationReq.setDelegatorAddress(delegator.MarshalPB());
            const result = yield this.staticCallAsync('CheckDelegation', checkDelegationReq, new dposv3_pb_1.CheckDelegationResponse());
            return {
                amount: result.getAmount() ? (0, big_uint_1.unmarshalBigUIntPB)(result.getAmount()) : new bn_js_1.default(0),
                weightedAmount: result.getWeightedAmount()
                    ? (0, big_uint_1.unmarshalBigUIntPB)(result.getWeightedAmount())
                    : new bn_js_1.default(0),
                delegationsArray: result.getDelegationsList().map(this.getDelegation)
            };
        });
    }
    registerCandidateAsync(pubKey, fee, name, description, website, tier) {
        const registerCandidateRequest = new dposv3_pb_1.RegisterCandidateRequestV3();
        registerCandidateRequest.setPubKey(pubKey);
        registerCandidateRequest.setFee(fee.toNumber());
        registerCandidateRequest.setName(name);
        registerCandidateRequest.setDescription(description);
        registerCandidateRequest.setWebsite(website);
        registerCandidateRequest.setLocktimeTier(tier);
        return this.callAsync('RegisterCandidate', registerCandidateRequest);
    }
    unregisterCandidateAsync() {
        const unregisterCandidateRequest = new dposv3_pb_1.UnregisterCandidateRequestV3();
        return this.callAsync('UnregisterCandidate', unregisterCandidateRequest);
    }
    delegateAsync(validator, amount, tier, referrer) {
        const delegateRequest = new dposv3_pb_1.DelegateRequest();
        delegateRequest.setValidatorAddress(validator.MarshalPB());
        delegateRequest.setAmount((0, big_uint_1.marshalBigUIntPB)(amount));
        delegateRequest.setLocktimeTier(tier);
        if (referrer) {
            delegateRequest.setReferrer(referrer);
        }
        return this.callAsync('Delegate', delegateRequest);
    }
    redelegateAsync(oldValidator, validator, amount, index, referrer) {
        const redelegateRequest = new dposv3_pb_1.RedelegateRequest();
        redelegateRequest.setFormerValidatorAddress(oldValidator.MarshalPB());
        redelegateRequest.setValidatorAddress(validator.MarshalPB());
        redelegateRequest.setAmount((0, big_uint_1.marshalBigUIntPB)(amount));
        redelegateRequest.setIndex(index);
        if (referrer) {
            redelegateRequest.setReferrer(referrer);
        }
        return this.callAsync('Redelegate', redelegateRequest);
    }
    consolidateDelegations(validator) {
        const req = new dposv3_pb_1.ConsolidateDelegationsRequest();
        req.setValidatorAddress(validator.MarshalPB());
        return this.callAsync('ConsolidateDelegations', req);
    }
    unbondAsync(validator, amount, index) {
        const unbondRequest = new dposv3_pb_1.UnbondRequest();
        unbondRequest.setValidatorAddress(validator.MarshalPB());
        unbondRequest.setAmount((0, big_uint_1.marshalBigUIntPB)(new bn_js_1.default(amount)));
        unbondRequest.setIndex(index);
        return this.callAsync('Unbond', unbondRequest);
    }
    claimDelegatorRewardsAsync() {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const req = new dposv3_pb_1.ClaimDelegatorRewardsRequest();
            const resp = new dposv3_pb_1.ClaimDelegatorRewardsResponse();
            yield this.callAsync('ClaimRewardsFromAllValidators', req, resp);
            return resp.getAmount() ? (0, big_uint_1.unmarshalBigUIntPB)(resp.getAmount()) : new bn_js_1.default(0);
        });
    }
    checkDelegatorRewardsAsync(delegator) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const req = new dposv3_pb_1.CheckDelegatorRewardsRequest();
            req.setDelegator(delegator.MarshalPB());
            const result = yield this.staticCallAsync('CheckRewardsFromAllValidators', req, new dposv3_pb_1.CheckDelegatorRewardsResponse());
            return result.getAmount() ? (0, big_uint_1.unmarshalBigUIntPB)(result.getAmount()) : new bn_js_1.default(0);
        });
    }
    getStateAsync() {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const req = new dposv3_pb_1.GetStateRequest();
            const res = yield this.staticCallAsync('GetState', req, new dposv3_pb_1.GetStateResponse());
            const state = res.getState();
            const params = state.getParams();
            const maxYearlyRewards = (0, big_uint_1.unmarshalBigUIntPB)(params.getMaxYearlyReward());
            const totalWeightedAmountStaked = (0, big_uint_1.unmarshalBigUIntPB)(state.getTotalValidatorDelegations());
            const minCandidateFee = params.getMinCandidateFee();
            return {
                maxYearlyRewards,
                totalWeightedAmountStaked,
                minCandidateFee
            };
        });
    }
    getDelegation(delegation) {
        return {
            validator: address_1.Address.UnmarshalPB(delegation.getValidator()),
            updateValidator: delegation.getUpdateValidator()
                ? address_1.Address.UnmarshalPB(delegation.getUpdateValidator())
                : undefined,
            delegator: address_1.Address.UnmarshalPB(delegation.getDelegator()),
            amount: delegation.getAmount() ? (0, big_uint_1.unmarshalBigUIntPB)(delegation.getAmount()) : new bn_js_1.default(0),
            updateAmount: delegation.getUpdateAmount()
                ? (0, big_uint_1.unmarshalBigUIntPB)(delegation.getUpdateAmount())
                : new bn_js_1.default(0),
            index: delegation.getIndex(),
            lockTime: delegation.getLockTime(),
            lockTimeTier: delegation.getLocktimeTier(),
            state: delegation.getState(),
            referrer: delegation.getReferrer()
        };
    }
    updateCandidateInfoAsync(name, description, website, maxReferralPercentage) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const updateCandidateInfo = new dposv3_pb_1.UpdateCandidateInfoRequestV3();
            updateCandidateInfo.setName(name);
            updateCandidateInfo.setDescription(description);
            updateCandidateInfo.setWebsite(website);
            updateCandidateInfo.setMaxReferralPercentage(maxReferralPercentage);
            return this.callAsync('UpdateCandidateInfo', updateCandidateInfo);
        });
    }
    changeFeeAsync(fee) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const changeFee = new dposv3_pb_1.ChangeCandidateFeeRequestV3();
            changeFee.setFee(fee);
            return this.callAsync('ChangeFee', changeFee);
        });
    }
    getDowntimeRecordAsync(validator) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const req = new dposv3_pb_1.DowntimeRecordRequest();
            req.setValidator(validator.MarshalPB());
            const res = yield this.staticCallAsync('DowntimeRecord', req, new dposv3_pb_1.DowntimeRecordResponse());
            return res.getDowntimeRecordsList();
        });
    }
    unjailAsync(validator) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const req = new dposv3_pb_1.UnjailRequest();
            if (validator) {
                req.setValidator(validator.MarshalPB());
            }
            yield this.callAsync('Unjail', req);
        });
    }
}
exports.DPOS3 = DPOS3;
//# sourceMappingURL=dpos3.js.map