"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.TransferGateway = void 0;
const tslib_1 = require("tslib");
const bn_js_1 = tslib_1.__importDefault(require("bn.js"));
const contract_1 = require("../contract");
const address_1 = require("../address");
const transfer_gateway_pb_1 = require("../proto/transfer_gateway_pb");
const big_uint_1 = require("../big-uint");
const crypto_utils_1 = require("../crypto-utils");
class TransferGateway extends contract_1.Contract {
    static createAsync(client, callerAddr) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const contractAddr = yield client.getContractAddressAsync('gateway');
            if (!contractAddr) {
                throw Error('Failed to resolve contract address for TransferGateway');
            }
            return new TransferGateway({ contractAddr, callerAddr, client });
        });
    }
    constructor(params) {
        super(params);
        this.on(contract_1.Contract.EVENT, event => {
            if (!event.topics || event.topics.length === 0) {
                return;
            }
            if (event.topics[0] === TransferGateway.tokenWithdrawalSignedEventTopic) {
                const eventData = transfer_gateway_pb_1.TransferGatewayTokenWithdrawalSigned.deserializeBinary((0, crypto_utils_1.B64ToUint8Array)(event.data));
                let tokenId;
                let tokenAmount;
                let value;
                const tokenKind = eventData.getTokenKind();
                switch (tokenKind) {
                    case transfer_gateway_pb_1.TransferGatewayTokenKind.ERC721:
                        tokenId = (0, big_uint_1.unmarshalBigUIntPB)(eventData.getTokenId());
                        value = tokenId;
                        break;
                    case transfer_gateway_pb_1.TransferGatewayTokenKind.ERC721X:
                        tokenId = (0, big_uint_1.unmarshalBigUIntPB)(eventData.getTokenId());
                    // fallthrough
                    // tslint:disable-next-line: no-switch-case-fall-through
                    default:
                        tokenAmount = (0, big_uint_1.unmarshalBigUIntPB)(eventData.getTokenAmount());
                        value = tokenAmount;
                        break;
                }
                this.emit(TransferGateway.EVENT_TOKEN_WITHDRAWAL, {
                    tokenOwner: address_1.Address.UnmarshalPB(eventData.getTokenOwner()),
                    tokenContract: address_1.Address.UnmarshalPB(eventData.getTokenContract()),
                    tokenKind,
                    tokenId,
                    tokenAmount,
                    sig: eventData.getSig_asU8(),
                    value
                });
            }
            else if (event.topics[0] === TransferGateway.contractMappingConfirmedEventTopic) {
                const contractMappingConfirmed = transfer_gateway_pb_1.TransferGatewayContractMappingConfirmed.deserializeBinary((0, crypto_utils_1.B64ToUint8Array)(event.data));
                this.emit(TransferGateway.EVENT_CONTRACT_MAPPING_CONFIRMED, {
                    foreignContract: address_1.Address.UnmarshalPB(contractMappingConfirmed.getForeignContract()),
                    localContract: address_1.Address.UnmarshalPB(contractMappingConfirmed.getLocalContract())
                });
            }
        });
    }
    /**
     * Adds a contract mapping to the DAppChain Gateway using gatway owner signature.
     * A contract mapping associates a token contract on the DAppChain with it's counterpart on Ethereum.
     */
    addAuthorizedContractMappingAsync(params) {
        const { foreignContract, localContract } = params;
        const mappingContractRequest = new transfer_gateway_pb_1.TransferGatewayAddContractMappingRequest();
        mappingContractRequest.setForeignContract(foreignContract.MarshalPB());
        mappingContractRequest.setLocalContract(localContract.MarshalPB());
        return this.callAsync('AddAuthorizedContractMapping', mappingContractRequest);
    }
    /**
     * Adds a contract mapping to the DAppChain Gateway.
     * A contract mapping associates a token contract on the DAppChain with it's counterpart on Ethereum.
     */
    addContractMappingAsync(params) {
        const { foreignContract, localContract, foreignContractCreatorSig, foreignContractCreatorTxHash } = params;
        const mappingContractRequest = new transfer_gateway_pb_1.TransferGatewayAddContractMappingRequest();
        mappingContractRequest.setForeignContract(foreignContract.MarshalPB());
        mappingContractRequest.setLocalContract(localContract.MarshalPB());
        mappingContractRequest.setForeignContractCreatorSig(foreignContractCreatorSig);
        mappingContractRequest.setForeignContractTxHash(foreignContractCreatorTxHash);
        return this.callAsync('AddContractMapping', mappingContractRequest);
    }
    /**
     * Fetches contract mappings from the DAppChain Gateway.
     *
     * @returns An object containing both confirmed & pending contract mappings.
     */
    listContractMappingsAsync() {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const response = yield this.staticCallAsync('ListContractMapping', new transfer_gateway_pb_1.TransferGatewayListContractMappingRequest(), new transfer_gateway_pb_1.TransferGatewayListContractMappingResponse());
            const confirmed = response.getConfimedMappingsList().map(mapping => ({
                from: address_1.Address.UnmarshalPB(mapping.getFrom()),
                to: address_1.Address.UnmarshalPB(mapping.getTo())
            }));
            const pending = response.getConfimedMappingsList().map(mapping => ({
                from: address_1.Address.UnmarshalPB(mapping.getFrom()),
                to: address_1.Address.UnmarshalPB(mapping.getTo())
            }));
            return {
                confirmed,
                pending
            };
        });
    }
    /**
     * Sends a request to the DAppChain Gateway to begin withdrawal of an ERC721 token from the
     * current DAppChain account to an Ethereum account.
     * @param tokenId ERC721 token ID.
     * @param tokenContract DAppChain address of ERC721 contract.
     * @param recipient Ethereum address of the account the token should be withdrawn to, if this is
     *                  omitted the Gateway will attempt to use the Address Mapper to retrieve the
     *                  address of the Ethereum account mapped to the current DAppChain account.
     * @returns A promise that will be resolved when the DAppChain Gateway has accepted the withdrawal
     *          request.
     */
    withdrawERC721Async(tokenId, tokenContract, recipient) {
        const req = new transfer_gateway_pb_1.TransferGatewayWithdrawTokenRequest();
        req.setTokenKind(transfer_gateway_pb_1.TransferGatewayTokenKind.ERC721);
        req.setTokenId((0, big_uint_1.marshalBigUIntPB)(tokenId));
        req.setTokenContract(tokenContract.MarshalPB());
        if (recipient) {
            req.setRecipient(recipient.MarshalPB());
        }
        return this.callAsync('WithdrawToken', req);
    }
    /**
     * Sends a request to the DAppChain Gateway to begin withdrawal of ERC721X tokens from the current
     * DAppChain account to an Ethereum account.
     * @param tokenId ERC721X token ID.
     * @param amount Amount of tokenId to withdraw.
     * @param tokenContract DAppChain address of ERC721X contract.
     * @param recipient Ethereum address of the account the token should be withdrawn to, if this is
     *                  omitted the Gateway will attempt to use the Address Mapper to retrieve the
     *                  address of the Ethereum account mapped to the current DAppChain account.
     * @returns A promise that will be resolved when the DAppChain Gateway has accepted the withdrawal
     *          request.
     */
    withdrawERC721XAsync(tokenId, amount, tokenContract, recipient) {
        const req = new transfer_gateway_pb_1.TransferGatewayWithdrawTokenRequest();
        req.setTokenKind(transfer_gateway_pb_1.TransferGatewayTokenKind.ERC721X);
        req.setTokenId((0, big_uint_1.marshalBigUIntPB)(tokenId));
        req.setTokenAmount((0, big_uint_1.marshalBigUIntPB)(amount));
        req.setTokenContract(tokenContract.MarshalPB());
        if (recipient) {
            req.setRecipient(recipient.MarshalPB());
        }
        return this.callAsync('WithdrawToken', req);
    }
    /**
     * Sends a request to the DAppChain Gateway to begin withdrawal ERC20 tokens from the current
     * DAppChain account to an Ethereum account.
     * @param amount Amount to withdraw.
     * @param tokenContract DAppChain address of ERC20 contract.
     * @param recipient Ethereum address of the account the token should be withdrawn to, if this is
     *                  omitted the Gateway will attempt to use the Address Mapper to retrieve the
     *                  address of the Ethereum account mapped to the current DAppChain account.
     * @returns A promise that will be resolved when the DAppChain Gateway has accepted the withdrawal
     *          request.
     */
    withdrawERC20Async(amount, tokenContract, recipient) {
        const req = new transfer_gateway_pb_1.TransferGatewayWithdrawTokenRequest();
        req.setTokenKind(transfer_gateway_pb_1.TransferGatewayTokenKind.ERC20);
        req.setTokenAmount((0, big_uint_1.marshalBigUIntPB)(amount));
        req.setTokenContract(tokenContract.MarshalPB());
        if (recipient) {
            req.setRecipient(recipient.MarshalPB());
        }
        return this.callAsync('WithdrawToken', req);
    }
    /**
     * Sends a request to the DAppChain Gateway to begin withdrawal of ETH from the current
     * DAppChain account to an Ethereum account.
     * @param amount Amount to withdraw.
     * @param ethereumGateway Ethereum address of Ethereum Gateway.
     * @param recipient Ethereum address of the account the token should be withdrawn to, if this is
     *                  omitted the Gateway will attempt to use the Address Mapper to retrieve the
     *                  address of the Ethereum account mapped to the current DAppChain account.
     * @returns A promise that will be resolved when the DAppChain Gateway has accepted the withdrawal
     *          request.
     */
    withdrawETHAsync(amount, ethereumGateway, recipient) {
        const req = new transfer_gateway_pb_1.TransferGatewayWithdrawETHRequest();
        req.setAmount((0, big_uint_1.marshalBigUIntPB)(amount));
        req.setMainnetGateway(ethereumGateway.MarshalPB());
        if (recipient) {
            req.setRecipient(recipient.MarshalPB());
        }
        return this.callAsync('WithdrawETH', req);
    }
    /**
     * Sends a request to the DAppChain Gateway to begin withdrawal TRX tokens from the current
     * DAppChain account to an Tron account.
     * @param amount Amount to withdraw.
     * @param tokenContract DAppChain address of Tron coin contract.
     * @param recipient Tron address of the account in hex string where the token should be withdrawn to, if this is
     *                  omitted the Gateway will attempt to use the Address Mapper to retrieve the
     *                  address of the Tron account mapped to the current DAppChain account.
     * @returns A promise that will be resolved when the DAppChain Gateway has accepted the withdrawal
     *          request.
     */
    withdrawTRXAsync(amount, tokenContract, recipient) {
        const req = new transfer_gateway_pb_1.TransferGatewayWithdrawTokenRequest();
        req.setTokenKind(transfer_gateway_pb_1.TransferGatewayTokenKind.TRX);
        req.setTokenAmount((0, big_uint_1.marshalBigUIntPB)(amount));
        req.setTokenContract(tokenContract.MarshalPB());
        if (recipient) {
            req.setRecipient(recipient.MarshalPB());
        }
        return this.callAsync('WithdrawToken', req);
    }
    /**
     * Sends a request to the DAppChain Gateway to begin withdrawal TRC20 tokens from the current
     * DAppChain account to an Tron account.
     * @param amount Amount to withdraw.
     * @param tokenContract DAppChain address of Tron TRC20 contract.
     * @param recipient Tron address of the account in hex string where the token should be withdrawn to, if this is
     *                  omitted the Gateway will attempt to use the Address Mapper to retrieve the
     *                  address of the Tron account mapped to the current DAppChain account.
     * @returns A promise that will be resolved when the DAppChain Gateway has accepted the withdrawal
     *          request.
     */
    withdrawTRC20Async(amount, tokenContract, recipient) {
        const req = new transfer_gateway_pb_1.TransferGatewayWithdrawTokenRequest();
        req.setTokenKind(transfer_gateway_pb_1.TransferGatewayTokenKind.TRC20);
        req.setTokenAmount((0, big_uint_1.marshalBigUIntPB)(amount));
        req.setTokenContract(tokenContract.MarshalPB());
        if (recipient) {
            req.setRecipient(recipient.MarshalPB());
        }
        return this.callAsync('WithdrawToken', req);
    }
    /**
     * Retrieves the current withdrawal receipt (if any) for the given DAppChain account.
     * Withdrawal receipts are created by calling one of the withdraw methods.
     * @param owner DAppChain address of a user account.
     * @returns A promise that will be resolved with the withdrawal receipt, or null if no withdrawal
     *          receipt exists (this indicates there's no withdrawal from the specified account
     *          currently in progress).
     */
    withdrawalReceiptAsync(owner) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const tgWithdrawReceiptReq = new transfer_gateway_pb_1.TransferGatewayWithdrawalReceiptRequest();
            tgWithdrawReceiptReq.setOwner(owner.MarshalPB());
            const result = yield this.staticCallAsync('WithdrawalReceipt', tgWithdrawReceiptReq, new transfer_gateway_pb_1.TransferGatewayWithdrawalReceiptResponse());
            const receipt = result.getReceipt();
            return receipt ? unmarshalWithdrawalReceipt(receipt) : null;
        });
    }
    /**
     * Attempt to transfer tokens that originated from the specified Ethereum contract, and that have
     * been deposited to the Ethereum Gateway, but haven't yet been received by the depositors on the
     * DAppChain because of a missing identity or contract mapping. This method can only be called by
     * the creator of the specified token contract, or the Gateway owner.
     *
     * @param tokenContract token contract to reclaim the tokens
     */
    reclaimContractTokensAsync(tokenContract) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const req = new transfer_gateway_pb_1.TransferGatewayReclaimContractTokensRequest();
            req.setTokenContract(tokenContract.MarshalPB());
            return this.callAsync('ReclaimContractTokens', req);
        });
    }
    getUnclaimedTokensAsync(owner) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const req = new transfer_gateway_pb_1.TransferGatewayGetUnclaimedTokensRequest();
            req.setOwner(owner.MarshalPB());
            const result = yield this.staticCallAsync('GetUnclaimedTokens', req, new transfer_gateway_pb_1.TransferGatewayGetUnclaimedTokensResponse());
            const unclaimedTokens = result.getUnclaimedTokensList();
            let tokens = [];
            for (let token of unclaimedTokens) {
                const tokenKind = token.getTokenKind();
                let tokenIds = [];
                let tokenAmounts = [];
                if (tokenKind === transfer_gateway_pb_1.TransferGatewayTokenKind.ERC721) {
                    // Set only the tokenId for ERC721
                    for (let amt of token.getAmountsList()) {
                        tokenIds.push((0, big_uint_1.unmarshalBigUIntPB)(amt.getTokenId()));
                    }
                }
                else if (tokenKind === transfer_gateway_pb_1.TransferGatewayTokenKind.ERC721X) {
                    // Set both the tokenId and the tokenAmounts for ERC721x
                    for (let amt of token.getAmountsList()) {
                        tokenIds.push((0, big_uint_1.unmarshalBigUIntPB)(amt.getTokenId()));
                        tokenAmounts.push((0, big_uint_1.unmarshalBigUIntPB)(amt.getTokenAmount()));
                    }
                }
                else {
                    // Set only amount for all other cases
                    for (let amt of token.getAmountsList()) {
                        tokenAmounts.push((0, big_uint_1.unmarshalBigUIntPB)(amt.getTokenAmount()));
                    }
                }
                tokens.push({
                    tokenContract: address_1.Address.UnmarshalPB(token.getTokenContract()),
                    tokenKind: tokenKind,
                    tokenAmounts: tokenAmounts,
                    tokenIds: tokenIds
                });
            }
            return tokens;
        });
    }
    /**
     * Attempt to transfer any tokens that the caller may have deposited into the Ethereum Gateway
     * but hasn't yet received from the DAppChain Gateway because of a missing identity or contract
     * mapping.
     *
     * @param depositors Optional list of DAppChain accounts to reclaim tokens for, when set tokens
     *                   will be reclaimed for the specified accounts instead of the caller's account.
     *                   NOTE: Only the Gateway owner is authorized to provide a list of accounts.
     */
    reclaimDepositorTokensAsync(depositors) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const req = new transfer_gateway_pb_1.TransferGatewayReclaimDepositorTokensRequest();
            if (depositors && depositors.length > 0) {
                req.setDepositorsList(depositors.map((address) => address.MarshalPB()));
            }
            return this.callAsync('ReclaimDepositorTokens', req);
        });
    }
    /**
     * Retrieves the information stored by the Gateway for the given DAppChain account.
     * @param owner DAppChain address of a user account.
     * @returns A promise that will be resolved with the local account info.
     */
    getLocalAccountInfoAsync(owner) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const req = new transfer_gateway_pb_1.TransferGatewayGetLocalAccountInfoRequest();
            req.setOwner(owner.MarshalPB());
            const resp = yield this.staticCallAsync('GetLocalAccountInfo', req, new transfer_gateway_pb_1.TransferGatewayGetLocalAccountInfoResponse());
            const receipt = resp.getWithdrawalReceipt();
            const amount = resp.getTotalWithdrawalAmount();
            return {
                withdrawalReceipt: receipt ? unmarshalWithdrawalReceipt(receipt) : null,
                totalWithdrawalAmount: amount ? (0, big_uint_1.unmarshalBigUIntPB)(amount) : new bn_js_1.default(0),
                lastWithdrawalLimitResetTime: resp.getLastWithdrawalLimitResetTime()
            };
        });
    }
    /**
     * Looks up the contract mapping for the given contract address.
     * @param from Contract address.
     * @returns null if no mapping was found for the given contract, otherwise an object that contains
     *          the address of the contract the `from` contract is mapped to, and the current status of
     *          the mapping (pending or confirmed).
     */
    getContractMappingAsync(from) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const request = new transfer_gateway_pb_1.TransferGatewayGetContractMappingRequest();
            request.setFrom(from.MarshalPB());
            const response = yield this.staticCallAsync('GetContractMapping', request, new transfer_gateway_pb_1.TransferGatewayGetContractMappingResponse());
            if (!response.getFound()) {
                return null;
            }
            return {
                to: address_1.Address.UnmarshalPB(response.getMappedAddress()),
                pending: response.getIsPending()
            };
        });
    }
    /**
     * Retrieves the current transfer gateway state.
     * @returns A promise that will be resolved with the state info.
     */
    getStateAsync() {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const response = yield this.staticCallAsync('GetState', new transfer_gateway_pb_1.TransferGatewayStateRequest(), new transfer_gateway_pb_1.TransferGatewayStateResponse());
            const state = response.getState();
            if (!state) {
                return {
                    maxTotalDailyWithdrawalAmount: new bn_js_1.default(0),
                    maxPerAccountDailyWithdrawalAmount: new bn_js_1.default(0),
                    lastWithdrawalLimitResetTime: new Date(0),
                    totalWithdrawalAmount: new bn_js_1.default(0)
                };
            }
            const maxTotalDailyWithdrawalAmount = state.getMaxTotalDailyWithdrawalAmount();
            const maxPerAccountDailyWithdrawalAmount = state.getMaxPerAccountDailyWithdrawalAmount();
            const lastWithdrawalLimitResetTime = state.getLastWithdrawalLimitResetTime();
            const totalWithdrawalAmount = state.getTotalWithdrawalAmount();
            return {
                maxTotalDailyWithdrawalAmount: maxTotalDailyWithdrawalAmount
                    ? (0, big_uint_1.unmarshalBigUIntPB)(maxTotalDailyWithdrawalAmount)
                    : new bn_js_1.default(0),
                maxPerAccountDailyWithdrawalAmount: maxPerAccountDailyWithdrawalAmount
                    ? (0, big_uint_1.unmarshalBigUIntPB)(maxPerAccountDailyWithdrawalAmount)
                    : new bn_js_1.default(0),
                lastWithdrawalLimitResetTime: lastWithdrawalLimitResetTime
                    ? new Date(lastWithdrawalLimitResetTime * 1000)
                    : new Date(0),
                totalWithdrawalAmount: totalWithdrawalAmount
                    ? (0, big_uint_1.unmarshalBigUIntPB)(totalWithdrawalAmount)
                    : new bn_js_1.default(0)
            };
        });
    }
}
exports.TransferGateway = TransferGateway;
// LoomJS user events
TransferGateway.EVENT_TOKEN_WITHDRAWAL = 'event_token_withdrawal';
TransferGateway.EVENT_CONTRACT_MAPPING_CONFIRMED = 'event_contract_mapping_confirmed';
// Events from Loomchain
TransferGateway.tokenWithdrawalSignedEventTopic = 'event:TokenWithdrawalSigned';
TransferGateway.contractMappingConfirmedEventTopic = 'event:ContractMappingConfirmed';
function unmarshalWithdrawalReceipt(receipt) {
    let tokenId;
    let tokenAmount;
    let value;
    const tokenKind = receipt.getTokenKind();
    switch (tokenKind) {
        case transfer_gateway_pb_1.TransferGatewayTokenKind.ERC721:
            tokenId = (0, big_uint_1.unmarshalBigUIntPB)(receipt.getTokenId());
            value = tokenId;
            break;
        case transfer_gateway_pb_1.TransferGatewayTokenKind.ERC721X:
            tokenId = (0, big_uint_1.unmarshalBigUIntPB)(receipt.getTokenId());
        // fallthrough
        // tslint:disable-next-line: no-switch-case-fall-through
        default:
            tokenAmount = (0, big_uint_1.unmarshalBigUIntPB)(receipt.getTokenAmount());
            value = tokenAmount;
            break;
    }
    return {
        tokenOwner: address_1.Address.UnmarshalPB(receipt.getTokenOwner()),
        tokenContract: receipt.getTokenContract()
            ? address_1.Address.UnmarshalPB(receipt.getTokenContract())
            : undefined,
        tokenKind,
        tokenId,
        tokenAmount,
        withdrawalNonce: new bn_js_1.default(receipt.getWithdrawalNonce()),
        oracleSignature: receipt.getOracleSignature_asU8(),
        txHash: receipt.getTxHash_asU8(),
        txStatus: receipt.getTxStatus(),
        value
    };
}
//# sourceMappingURL=transfer-gateway.js.map