"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.UserDeployerWhitelist = void 0;
const tslib_1 = require("tslib");
const bn_js_1 = tslib_1.__importDefault(require("bn.js"));
const contract_1 = require("../contract");
const address_1 = require("../address");
const user_deployer_whitelist_pb_1 = require("../proto/user_deployer_whitelist_pb");
const big_uint_1 = require("../big-uint");
/**
 * Provides self-service deployer account management for users that wish to deploy EVM contracts
 * to the DAppChain.
 */
class UserDeployerWhitelist extends contract_1.Contract {
    static createAsync(client, callerAddr) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const contractAddr = yield client.getContractAddressAsync('user-deployer-whitelist');
            if (!contractAddr) {
                throw Error('Failed to resolve contract address');
            }
            return new UserDeployerWhitelist({ contractAddr, callerAddr, client });
        });
    }
    constructor(params) {
        super(params);
    }
    /**
     * Authorizes an account to deploy EVM contracts on behalf of the caller. In order to authorize
     * an account the caller must approve the contract to withdraw the fee for whitelisting
     * (charged in LOOM coin) before calling this method, the fee will be deducted from the caller if
     * the requested account is successfuly authorized.
     *
     * @param deployer Deployer account address.
     * @param tierId
     */
    addDeployerAsync(deployer, tierId) {
        const req = new user_deployer_whitelist_pb_1.WhitelistUserDeployerRequest();
        req.setDeployerAddr(deployer.MarshalPB());
        req.setTierId(tierId ? tierId : user_deployer_whitelist_pb_1.TierID.DEFAULT);
        return this.callAsync('AddUserDeployer', req);
    }
    /**
     * Removes whitelisted deployer
     * @param deployer Deployer account address.
     */
    removeDeployerAsync(deployer) {
        const req = new user_deployer_whitelist_pb_1.RemoveUserDeployerRequest();
        req.setDeployerAddr(deployer.MarshalPB());
        return this.callAsync('RemoveUserDeployer', req);
    }
    /**
     * @param user User account address.
     * @returns List of accounts authorized to deploy EVM contracts on behalf of the specified user.
     */
    getDeployersAsync(user) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const req = new user_deployer_whitelist_pb_1.GetUserDeployersRequest();
            req.setUserAddr(user.MarshalPB());
            const result = yield this.staticCallAsync('GetUserDeployers', req, new user_deployer_whitelist_pb_1.GetUserDeployersResponse());
            return result.getDeployersList().map(userDeployerState => ({
                address: address_1.Address.UnmarshalPB(userDeployerState.getAddress()),
                contracts: userDeployerState.getContractsList().map(deployerContract => ({
                    address: address_1.Address.UnmarshalPB(deployerContract.getContractAddress())
                })),
                tierId: userDeployerState.getTierId(),
                inactive: userDeployerState.getInactive()
            }));
        });
    }
    /**
     * @param deployer Deployer account address.
     * @returns Array of EVM contracts deployed by a particular deployer account.
     */
    getDeployedContractsAsync(deployer) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const req = new user_deployer_whitelist_pb_1.GetDeployedContractsRequest();
            req.setDeployerAddr(deployer.MarshalPB());
            const result = yield this.staticCallAsync('GetDeployedContracts', req, new user_deployer_whitelist_pb_1.GetDeployedContractsResponse());
            return result.getContractAddressesList().map(deployerContract => ({
                address: address_1.Address.UnmarshalPB(deployerContract.getContractAddress())
            }));
        });
    }
    /**
     * @param tierId ID of tier.
     * @returns Tier details.
     */
    getTierInfoAsync(tierId) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const req = new user_deployer_whitelist_pb_1.GetTierInfoRequest();
            req.setId(tierId);
            const result = yield this.staticCallAsync('GetTierInfo', req, new user_deployer_whitelist_pb_1.GetTierInfoResponse());
            return {
                tierId: result.getTier().getTierId(),
                fee: (0, big_uint_1.unmarshalBigUIntPB)(result.getTier().getFee()),
                name: result.getTier().getName()
            };
        });
    }
    /**
     * Set tier details, can only be called by the UserDeployerWhitelist contract owner.
     *
     * @param tier Tier details.
     */
    setTierInfoAsync(tier) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const req = new user_deployer_whitelist_pb_1.SetTierInfoRequest();
            if (tier.fee.cmp(new bn_js_1.default(0)) <= 0) {
                throw Error('fee must be greater than zero');
            }
            req.setFee((0, big_uint_1.marshalBigUIntPB)(tier.fee));
            req.setId(tier.tierId);
            req.setName(tier.name);
            return this.callAsync('SetTierInfo', req);
        });
    }
}
exports.UserDeployerWhitelist = UserDeployerWhitelist;
//# sourceMappingURL=user-deployer-whitelist.js.map