"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.DPOSUser = void 0;
const tslib_1 = require("tslib");
const debug_1 = tslib_1.__importDefault(require("debug"));
const ethers_1 = require("ethers");
const _1 = require(".");
const contracts_1 = require("./contracts");
const helpers_1 = require("./helpers");
const gateway_user_1 = require("./gateway-user");
const solidity_helpers_1 = require("./solidity-helpers");
const log = (0, debug_1.default)('dpos-user');
class DPOSUser extends gateway_user_1.GatewayUser {
    static createOfflineUserAsync(endpoint, privateKey, dappchainEndpoint, dappchainKey, chainId, gatewayAddress, version) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const provider = new ethers_1.ethers.providers.JsonRpcProvider(endpoint);
            const wallet = new ethers_1.ethers.Wallet(privateKey, provider);
            return DPOSUser.createUserAsync(wallet, dappchainEndpoint, dappchainKey, chainId, gatewayAddress, version);
        });
    }
    static createMetamaskUserAsync(web3, dappchainEndpoint, dappchainKey, chainId, gatewayAddress, version) {
        const wallet = (0, solidity_helpers_1.getMetamaskSigner)(web3.currentProvider);
        return DPOSUser.createUserAsync(wallet, dappchainEndpoint, dappchainKey, chainId, gatewayAddress, version);
    }
    static createEthSignMetamaskUserAsync(web3, dappchainEndpoint, chainId, gatewayAddress, version) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const gatewayUser = yield gateway_user_1.GatewayUser.createEthSignMetamaskGatewayUserAsync({
                web3,
                dappchainEndpoint,
                chainId,
                gatewayAddress,
                version: version ? version : 1
            });
            const dappchainEthAddress = new _1.Address('eth', _1.LocalAddress.fromHexString(gatewayUser.ethAddress));
            const dappchainDPOS = yield contracts_1.DPOS2.createAsync(gatewayUser.client, dappchainEthAddress);
            log('Connected to dappchain DPOS Contract');
            return new DPOSUser(gatewayUser.wallet, gatewayUser.client, gatewayUser.loomAddress, gatewayUser.ethAddress, gatewayUser.ethereumGateway, gatewayUser.ethereumLoom, gatewayUser.ethereumVMC, gatewayUser.dappchainGateway, gatewayUser.dappchainLoom, dappchainDPOS, null, version);
        });
    }
    static createUserAsync(wallet, dappchainEndpoint, dappchainKey, chainId, gatewayAddress, version) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const gatewayUser = yield gateway_user_1.GatewayUser.createGatewayUserAsync({
                wallet,
                dappchainEndpoint,
                dappchainPrivateKey: dappchainKey,
                chainId,
                gatewayAddress,
                version: version ? version : 1
            });
            const { client, address } = (0, helpers_1.createDefaultClient)(dappchainKey, dappchainEndpoint, chainId);
            const dappchainDPOS = yield contracts_1.DPOS2.createAsync(client, address);
            log('Connected to dappchain DPOS Contract');
            return new DPOSUser(gatewayUser.wallet, gatewayUser.client, gatewayUser.loomAddress, gatewayUser.ethAddress, gatewayUser.ethereumGateway, gatewayUser.ethereumLoom, gatewayUser.ethereumVMC, gatewayUser.dappchainGateway, gatewayUser.dappchainLoom, dappchainDPOS, gatewayUser.addressMapper, version);
        });
    }
    constructor(wallet, client, loomAddress, ethAddress, gateway, loomToken, vmc, dappchainGateway, dappchainLoom, dappchainDPOS, dappchainMapper, version = 1) {
        super({
            wallet,
            client,
            loomAddress,
            ethAddress,
            gateway,
            loomToken,
            vmc,
            dappchainGateway,
            dappchainLoom,
            addressMapper: dappchainMapper,
            version
        });
        this._dappchainDPOS = dappchainDPOS;
    }
    get dappchainDPOS() {
        return this._dappchainDPOS;
    }
    listValidatorsAsync() {
        return this._dappchainDPOS.getValidatorsAsync();
    }
    listCandidatesAsync() {
        return this._dappchainDPOS.getCandidatesAsync();
    }
    listAllDelegationsAsync() {
        return this._dappchainDPOS.getAllDelegations();
    }
    listDelegationsAsync(candidate) {
        const address = this.prefixAddress(candidate);
        return this._dappchainDPOS.getDelegations(address);
    }
    listDelegatorDelegations(delegator) {
        const address = delegator ? this.prefixAddress(delegator) : this.loomAddress;
        return this._dappchainDPOS.checkDelegatorDelegations(address);
    }
    getTimeUntilElectionsAsync() {
        return this._dappchainDPOS.getTimeUntilElectionAsync();
    }
    /**
     * Delegates an amount of LOOM tokens to a candidate/validator
     *
     * @param candidate The candidate's hex address
     * @param amount The amount delegated
     */
    delegateAsync(candidate, amount, tier) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const address = this.prefixAddress(candidate);
            yield this.dappchainLoom.approveAsync(this._dappchainDPOS.address, amount);
            return this._dappchainDPOS.delegateAsync(address, amount, tier);
        });
    }
    /**
     * Redelegates an amount of LOOM tokens from a validator to another
     *
     * @param formerValidator The candidate's hex address
     * @param newValidator The candidate's hex address
     * @param amount The amount delegated
     */
    redelegateAsync(formerValidator, validator, amount) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const validatorAddress = this.prefixAddress(validator);
            const formerValidatorAddress = this.prefixAddress(formerValidator);
            return this._dappchainDPOS.redelegateAsync(formerValidatorAddress, validatorAddress, amount);
        });
    }
    /**
     * Undelegates an amount of LOOM tokens from a candidate/validator
     *
     * @param candidate The candidate's hex address
     * @param amount The amount to undelegate
     */
    undelegateAsync(candidate, amount) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const address = this.prefixAddress(candidate);
            yield this._dappchainDPOS.unbondAsync(address, amount);
        });
    }
    claimDelegationsAsync(withdrawalAddress) {
        const address = withdrawalAddress ? this.prefixAddress(withdrawalAddress) : this.loomAddress;
        return this._dappchainDPOS.claimDistributionAsync(address);
    }
    /**
     * Returns the stake a delegator has delegated to a candidate/validator
     *
     * @param validator The validator's hex address
     * @param delegator The delegator's hex address
     */
    checkDelegationsAsync(validator, delegator) {
        const validatorAddress = this.prefixAddress(validator);
        const delegatorAddress = delegator ? this.prefixAddress(delegator) : this.loomAddress;
        return this._dappchainDPOS.checkDelegationAsync(validatorAddress, delegatorAddress);
    }
    /**
     * Returns the total stake a delegator has delegated to all validators
     *
     * @param delegator The delegator's hex address. If not supplied, will use the current account as a delegator.
     */
    getTotalDelegationAsync(delegator) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const delegatorAddress = delegator ? this.prefixAddress(delegator) : this.loomAddress;
            return this._dappchainDPOS.totalDelegationAsync(delegatorAddress);
        });
    }
    checkRewardsAsync(owner) {
        const address = owner ? this.prefixAddress(owner) : this.loomAddress;
        return this._dappchainDPOS.checkDistributionAsync(address);
    }
}
exports.DPOSUser = DPOSUser;
//# sourceMappingURL=dpos-user.js.map