"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.createEthereumGatewayAsync = exports.EthereumGatewayV2 = exports.EthereumGatewayV1 = void 0;
const tslib_1 = require("tslib");
const debug_1 = tslib_1.__importDefault(require("debug"));
const ethers_1 = require("ethers");
const _1 = require(".");
const helpers_1 = require("./helpers");
const transfer_gateway_pb_js_1 = require("./proto/transfer_gateway_pb.js");
const ValidatorManagerV2__factory_1 = require("./mainnet-contracts/factories/ValidatorManagerV2__factory");
const EthereumGatewayV2__factory_1 = require("./mainnet-contracts/factories/EthereumGatewayV2__factory");
const EthereumGatewayV1__factory_1 = require("./mainnet-contracts/factories/EthereumGatewayV1__factory");
const log = (0, debug_1.default)('loom.ethereum');
class EthereumGatewayV1 {
    constructor(contract) {
        this.contract = contract;
        this.version = 1;
    }
    withdrawAsync(receipt, overrides) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const signature = _1.CryptoUtils.bytesToHexAddr(receipt.oracleSignature);
            let result;
            switch (receipt.tokenKind) {
                case transfer_gateway_pb_js_1.TransferGatewayTokenKind.ETH:
                    result = this.contract.functions.withdrawETH(receipt.tokenAmount.toString(), signature, overrides);
                    break;
                case transfer_gateway_pb_js_1.TransferGatewayTokenKind.LOOMCOIN:
                case transfer_gateway_pb_js_1.TransferGatewayTokenKind.ERC20:
                    result = this.contract.functions.withdrawERC20(receipt.tokenAmount.toString(), signature, receipt.tokenContract.local.toString(), overrides);
                    break;
                case transfer_gateway_pb_js_1.TransferGatewayTokenKind.ERC721:
                    result = this.contract.functions.withdrawERC721(receipt.tokenId.toString(), signature, receipt.tokenContract.local.toString(), overrides);
                    break;
                case transfer_gateway_pb_js_1.TransferGatewayTokenKind.ERC721X:
                    result = this.contract.functions.withdrawERC721X(receipt.tokenId.toString(), receipt.tokenAmount.toString(), signature, receipt.tokenContract.local.toString(), overrides);
                    break;
                default:
                    throw new Error('Unsupported token kind ' + receipt.tokenKind);
            }
            return result;
        });
    }
    depositERC20Async(amount, contractAddress, overrides) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            return this.contract.functions.depositERC20(amount.toString(), contractAddress, overrides);
        });
    }
    withSigner(signer) {
        return new EthereumGatewayV1(this.contract.connect(signer));
    }
}
exports.EthereumGatewayV1 = EthereumGatewayV1;
class EthereumGatewayV2 {
    constructor(contract, vmc) {
        this.contract = contract;
        this.vmc = vmc;
        this.version = 2;
    }
    withdrawAsync(receipt, overrides) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const validators = yield this.vmc.getValidators();
            const hash = createWithdrawalHash(receipt, this.contract.address);
            const { vs, rs, ss, valIndexes } = (0, helpers_1.parseSigs)(_1.CryptoUtils.bytesToHexAddr(receipt.oracleSignature), hash, validators);
            let result;
            switch (receipt.tokenKind) {
                case transfer_gateway_pb_js_1.TransferGatewayTokenKind.ETH:
                    result = this.contract.functions.withdrawETH(receipt.tokenAmount.toString(), valIndexes, vs, rs, ss, overrides);
                    break;
                case transfer_gateway_pb_js_1.TransferGatewayTokenKind.LOOMCOIN:
                case transfer_gateway_pb_js_1.TransferGatewayTokenKind.ERC20:
                    result = this.contract.functions.withdrawERC20(receipt.tokenAmount.toString(), receipt.tokenContract.local.toString(), valIndexes, vs, rs, ss, overrides);
                    break;
                case transfer_gateway_pb_js_1.TransferGatewayTokenKind.ERC721:
                    result = this.contract.functions.withdrawERC721(receipt.tokenId.toString(), receipt.tokenContract.local.toString(), valIndexes, vs, rs, ss, overrides);
                    break;
                case transfer_gateway_pb_js_1.TransferGatewayTokenKind.ERC721X:
                    result = this.contract.functions.withdrawERC721X(receipt.tokenId.toString(), receipt.tokenAmount.toString(), receipt.tokenContract.local.toString(), valIndexes, vs, rs, ss, overrides);
                    break;
                default:
                    throw new Error('Unsupported token kind ' + receipt.tokenKind);
            }
            return result;
        });
    }
    depositERC20Async(amount, contractAddress, overrides) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            return this.contract.functions.depositERC20(amount.toString(), contractAddress, overrides);
        });
    }
    withSigner(signer) {
        return new EthereumGatewayV2(this.contract.connect(signer), this.vmc);
    }
}
exports.EthereumGatewayV2 = EthereumGatewayV2;
/**
 * Creates an Ethereum Gateway contract wrapper.
 * @param version Ethereum Gateway contract version, must be 1, or 2.
 * @param address Ethereum Gateway address.
 * @param provider web3 provider.
 */
function createEthereumGatewayAsync(version, address, provider) {
    return tslib_1.__awaiter(this, void 0, void 0, function* () {
        const gatewayContract = EthereumGatewayV2__factory_1.EthereumGatewayV2__factory.connect(address, provider);
        switch (version) {
            case 2:
                const vmcAddress = yield gatewayContract.vmc();
                const vmcContract = ValidatorManagerV2__factory_1.ValidatorManagerV2__factory.connect(vmcAddress, provider);
                return new EthereumGatewayV2(gatewayContract, vmcContract);
            case 1:
                return new EthereumGatewayV1(EthereumGatewayV1__factory_1.EthereumGatewayV1__factory.connect(address, provider));
            default:
                throw new Error('Invalid Ethereum Gateway version: ' + version);
        }
    });
}
exports.createEthereumGatewayAsync = createEthereumGatewayAsync;
var MessagePrefix;
(function (MessagePrefix) {
    MessagePrefix["ETH"] = "\u000EWithdraw ETH:\n";
    MessagePrefix["ERC20"] = "\u0010Withdraw ERC20:\n";
    MessagePrefix["ERC721"] = "\u0011Withdraw ERC721:\n";
    MessagePrefix["ERC721X"] = "\u0012Withdraw ERC721X:\n";
})(MessagePrefix || (MessagePrefix = {}));
function createWithdrawalHash(receipt, gatewayAddress) {
    let prefix;
    let amountHashed;
    switch (receipt.tokenKind) {
        case transfer_gateway_pb_js_1.TransferGatewayTokenKind.ERC721:
            prefix = MessagePrefix.ERC721;
            amountHashed = ethers_1.ethers.utils.solidityKeccak256(['uint256', 'address'], [receipt.tokenId ? receipt.tokenId.toString() : 0, receipt.tokenContract.local.toString()]);
            break;
        case transfer_gateway_pb_js_1.TransferGatewayTokenKind.ERC721X:
            prefix = MessagePrefix.ERC721X;
            amountHashed = ethers_1.ethers.utils.solidityKeccak256(['uint256', 'uint256', 'address'], [
                receipt.tokenId ? receipt.tokenId.toString() : 0,
                receipt.tokenAmount ? receipt.tokenAmount.toString() : 0,
                receipt.tokenContract.local.toString()
            ]);
            break;
        case transfer_gateway_pb_js_1.TransferGatewayTokenKind.LOOMCOIN:
        case transfer_gateway_pb_js_1.TransferGatewayTokenKind.ERC20:
            prefix = MessagePrefix.ERC20;
            amountHashed = ethers_1.ethers.utils.solidityKeccak256(['uint256', 'address'], [
                receipt.tokenAmount ? receipt.tokenAmount.toString() : 0,
                receipt.tokenContract.local.toString()
            ]);
            break;
        case transfer_gateway_pb_js_1.TransferGatewayTokenKind.ETH:
            prefix = MessagePrefix.ETH;
            amountHashed = ethers_1.ethers.utils.solidityKeccak256(['uint256'], [receipt.tokenAmount ? receipt.tokenAmount.toString() : 0]);
            break;
        default:
            throw new Error('Unsupported token kind');
    }
    return ethers_1.ethers.utils.solidityKeccak256(['string', 'address', 'uint256', 'address', 'bytes32'], [
        prefix,
        receipt.tokenOwner.local.toString(),
        receipt.withdrawalNonce.toString(),
        gatewayAddress,
        amountHashed
    ]);
}
//# sourceMappingURL=ethereum-gateways.js.map