"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.EvmContract = void 0;
const tslib_1 = require("tslib");
const events_1 = tslib_1.__importDefault(require("events"));
const client_1 = require("./client");
const loom_pb_1 = require("./proto/loom_pb");
/**
 * The EvmContract class streamlines interaction with a contract that was
 * deployed on a Loom DAppChain EVM.
 * Each instance of this class is bound to a specific contract, and provides a simple way of calling
 * into and querying that contract.
 *
 * A contract instance can be used to listen to events emitted by the corresponding smart contract,
 * there is currently only one type of event. The event subscription API matches the NodeJS
 * EventEmitter API.
 */
class EvmContract extends events_1.default {
    /**
     * @param params Parameters.
     * @param params.contractAddr Address of a contract on the Loom DAppChain EVM.
     * @param params.callerAddr: Address of the caller, generated from the public key of the tx signer,
     *                           e.g. `new Address(client.chainId, LocalAddress.fromPublicKey(pubKey))`
     * @param params.client: Client to use to communicate with the contract.
     */
    constructor(params) {
        super();
        this._client = params.client;
        this.address = params.contractAddr;
        this.caller = params.callerAddr;
        const emitContractEvent = this._emitContractEvent.bind(this);
        this.on('newListener', (event) => {
            if (event === EvmContract.EVENT && this.listenerCount(event) === 0) {
                this._client.on(client_1.ClientEvent.Contract, emitContractEvent);
            }
        });
        this.on('removeListener', (event) => {
            if (event === EvmContract.EVENT && this.listenerCount(event) === 0) {
                this._client.removeListener(client_1.ClientEvent.Contract, emitContractEvent);
            }
        });
    }
    /**
     * Calls a contract method that mutates state.
     * The call into the contract is accomplished by committing a tx to the DAppChain.
     * @param args ABI encoded function signature and input paramters.
     * @returns A promise that will be resolved with return value (if any) of the contract method.
     */
    callAsync(args, output) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const ui8InData = Uint8Array.from(args);
            const callTx = new loom_pb_1.CallTx();
            callTx.setVmType(loom_pb_1.VMType.EVM);
            callTx.setInput(ui8InData);
            const msgTx = new loom_pb_1.MessageTx();
            msgTx.setFrom(this.caller.MarshalPB());
            msgTx.setTo(this.address.MarshalPB());
            msgTx.setData(callTx.serializeBinary());
            const tx = new loom_pb_1.Transaction();
            tx.setId(2);
            tx.setData(msgTx.serializeBinary());
            return this._client.commitTxAsync(tx);
        });
    }
    /**
     * Calls a method of a contract running on an EVM that doesn't mutate state.
     * This method is usually used to query the current contract state, it doesn't commit any txs.
     * @param args ABI encoded function signature and input paramters.
     * @returns A promise that will be resolved with the return value of the contract method.
     */
    staticCallAsync(args, output) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const ui8InData = Uint8Array.from(args);
            return this._client.queryAsync(this.address, ui8InData, loom_pb_1.VMType.EVM, this.caller);
        });
    }
    _emitContractEvent(event) {
        if (event.contractAddress.equals(this.address)) {
            this.emit(EvmContract.EVENT, event);
        }
    }
}
exports.EvmContract = EvmContract;
EvmContract.EVENT = 'event';
//# sourceMappingURL=evm-contract.js.map