"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.GatewayUser = void 0;
const tslib_1 = require("tslib");
const bn_js_1 = tslib_1.__importDefault(require("bn.js"));
const debug_1 = tslib_1.__importDefault(require("debug"));
const ethers_1 = require("ethers");
const _1 = require(".");
const contracts_1 = require("./contracts");
const helpers_1 = require("./helpers");
const solidity_helpers_1 = require("./solidity-helpers");
const crosschain_user_1 = require("./crosschain-user");
const log = (0, debug_1.default)('gateway-user');
const coinMultiplier = new bn_js_1.default(10).pow(new bn_js_1.default(18));
const ERC20__factory_1 = require("./mainnet-contracts/factories/ERC20__factory");
const EthereumGatewayV2__factory_1 = require("./mainnet-contracts/factories/EthereumGatewayV2__factory");
const EthereumGatewayV1__factory_1 = require("./mainnet-contracts/factories/EthereumGatewayV1__factory");
const ValidatorManagerV2__factory_1 = require("./mainnet-contracts/factories/ValidatorManagerV2__factory");
const ERC20Prefix = '\x10Withdraw ERC20:\n';
const V2_GATEWAYS = ['asia1', 'test-z-us-1', 'extdev-plasma-us1'];
/**
 * @deprecated Will be removed in loom-js v2.0.0
 */
class GatewayUser extends crosschain_user_1.CrossChainUser {
    static createGatewayOfflineUserAsync(params) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const provider = new ethers_1.ethers.providers.JsonRpcProvider(params.ethEndpoint);
            const wallet = new ethers_1.ethers.Wallet(params.ethereumPrivateKey, provider);
            return GatewayUser.createGatewayUserAsync(Object.assign({ wallet }, params));
        });
    }
    static createGatewayMetamaskUserAsync(params) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const wallet = (0, solidity_helpers_1.getMetamaskSigner)(params.web3.currentProvider);
            return GatewayUser.createGatewayUserAsync(Object.assign({ wallet }, params));
        });
    }
    static getContracts(wallet, gatewayAddress, version) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            let gateway;
            switch (version) {
                case 2:
                    gateway = EthereumGatewayV2__factory_1.EthereumGatewayV2__factory.connect(gatewayAddress, wallet);
                    break;
                case 1:
                    gateway = EthereumGatewayV1__factory_1.EthereumGatewayV1__factory.connect(gatewayAddress, wallet);
                    break;
                default:
                    throw new Error('Invalid Ethereum Gateway version ' + version);
            }
            const loomAddress = yield gateway.functions.loomAddress();
            const loomToken = ERC20__factory_1.ERC20__factory.connect(loomAddress, wallet);
            let vmc;
            if (version === 2) {
                const vmcAddress = yield gateway.functions.vmc();
                vmc = ValidatorManagerV2__factory_1.ValidatorManagerV2__factory.connect(vmcAddress, wallet);
            }
            return {
                gateway: gateway,
                loomToken: loomToken,
                vmc: vmc
            };
        });
    }
    static getGatewayVersion(endpoint, version) {
        // If no gateway version is provided, pick based on the chain URL prefix
        let retVersion = 1;
        if (typeof version === undefined) {
            const chainName = endpoint.split('.')[0];
            for (let chainPrefix of V2_GATEWAYS) {
                if (chainName.indexOf(chainPrefix) != -1) {
                    retVersion = 2;
                }
            }
        }
        return retVersion;
    }
    static createEthSignMetamaskGatewayUserAsync(params) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const gwVersion = GatewayUser.getGatewayVersion(params.dappchainEndpoint, params.version);
            let crosschain = yield crosschain_user_1.CrossChainUser.createEthSignMetamaskCrossChainUserAsync(params);
            const dappchainEthAddress = new _1.Address('eth', _1.LocalAddress.fromHexString(crosschain.ethAddress));
            const dappchainLoom = yield contracts_1.Coin.createAsync(crosschain.client, dappchainEthAddress);
            const dappchainGateway = yield contracts_1.LoomCoinTransferGateway.createAsync(crosschain.client, dappchainEthAddress);
            const { gateway, loomToken, vmc } = yield GatewayUser.getContracts(crosschain.wallet, params.gatewayAddress, params.version);
            return new GatewayUser({
                wallet: crosschain.wallet,
                client: crosschain.client,
                loomAddress: crosschain.loomAddress,
                ethAddress: crosschain.ethAddress,
                gateway,
                loomToken,
                vmc,
                dappchainGateway,
                dappchainLoom,
                version: gwVersion
            });
        });
    }
    static createGatewayUserAsync(params) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const gwVersion = GatewayUser.getGatewayVersion(params.dappchainEndpoint, params.version);
            let crosschain = yield crosschain_user_1.CrossChainUser.createCrossChainUserAsync(params);
            const dappchainLoom = yield contracts_1.Coin.createAsync(crosschain.client, crosschain.loomAddress);
            const dappchainGateway = yield contracts_1.LoomCoinTransferGateway.createAsync(crosschain.client, crosschain.loomAddress);
            const { gateway, loomToken, vmc } = yield GatewayUser.getContracts(crosschain.wallet, params.gatewayAddress, params.version);
            return new GatewayUser({
                wallet: crosschain.wallet,
                client: crosschain.client,
                loomAddress: crosschain.loomAddress,
                ethAddress: crosschain.ethAddress,
                gateway,
                loomToken,
                vmc,
                dappchainGateway,
                dappchainLoom,
                addressMapper: crosschain.addressMapper,
                version: gwVersion
            });
        });
    }
    constructor(params) {
        super(params);
        this._version = params.version;
        // Set ethereum contracts
        // @ts-ignore
        this._ethereumGateway = params.gateway;
        // @ts-ignore
        this._ethereumLoom = params.loomToken;
        // @ts-ignore
        this._ethereumVMC = params.vmc;
        // Set dappchain contracts
        this._dappchainGateway = params.dappchainGateway;
        this._dappchainLoom = params.dappchainLoom;
        this._dappchainGateway = params.dappchainGateway;
    }
    get ethereumVMC() {
        return this._ethereumVMC;
    }
    get ethereumGateway() {
        return this._ethereumGateway;
    }
    get ethereumLoom() {
        return this._ethereumLoom;
    }
    get dappchainLoom() {
        return this._dappchainLoom;
    }
    get dappchainGateway() {
        return this._dappchainGateway;
    }
    /**
     * Deposits funds from mainnet to the gateway
     */
    depositAsync(amount) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            let currentApproval = yield this._ethereumLoom.functions.allowance(yield this.ethAddress, this._ethereumGateway.address);
            let currentApprovalBN = new bn_js_1.default(currentApproval.toString());
            log('Current approval:', currentApproval);
            if (amount.gt(currentApprovalBN)) {
                let tx = yield this._ethereumLoom.functions.approve(this._ethereumGateway.address, amount.sub(currentApprovalBN).toString());
                yield tx.wait();
                log('Approved an extra', amount.sub(currentApprovalBN));
            }
            return this._ethereumGateway.functions.depositERC20(amount.toString(), this._ethereumLoom.address);
        });
    }
    /**
     * Withdraw funds from the gateway to mainnet
     */
    withdrawAsync(amount) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const sig = yield this.depositCoinToDAppChainGatewayAsync(amount);
            return this.withdrawCoinFromDAppChainGatewayAsync(amount, sig);
        });
    }
    resumeWithdrawalAsync() {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const receipt = yield this.getPendingWithdrawalReceiptAsync();
            if (receipt === null) {
                log('No pending receipt');
                return;
            }
            const signature = _1.CryptoUtils.bytesToHexAddr(receipt.oracleSignature);
            const amount = receipt.tokenAmount;
            return this.withdrawCoinFromDAppChainGatewayAsync(amount, signature);
        });
    }
    getPendingWithdrawalReceiptAsync() {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            return this._dappchainGateway.withdrawalReceiptAsync(this.loomAddress);
        });
    }
    /**
     * Retrieves the  DAppChain LoomCoin balance of a user
     * @param address The address to check the balance of. If not provided, it will check the user's balance
     */
    getDAppChainBalanceAsync(address) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const addr = address ? this.prefixAddress(address) : this.loomAddress;
            const balance = yield this._dappchainLoom.getBalanceOfAsync(addr);
            return balance;
        });
    }
    /**
     * Deposits an amount of LOOM tokens to the dappchain gateway and return a signature which can be used to withdraw the same amount from the mainnet gateway.
     *
     * @param amount The amount that will be deposited to the DAppChain Gateway (and will be possible to withdraw from the mainnet)
     */
    depositCoinToDAppChainGatewayAsync(amount) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            let pendingReceipt = yield this.getPendingWithdrawalReceiptAsync();
            let signature;
            if (pendingReceipt === null) {
                yield this._dappchainLoom.approveAsync(this._dappchainGateway.address, amount);
                const ethereumAddressStr = yield this.ethAddress;
                const ethereumAddress = _1.Address.fromString(`eth:${ethereumAddressStr}`);
                const _ethereumLoomCoinAddress = _1.Address.fromString(`eth:${this._ethereumLoom.address}`);
                yield this._dappchainGateway.withdrawLoomCoinAsync(amount, _ethereumLoomCoinAddress, ethereumAddress);
                log(`${amount.div(coinMultiplier).toString()} tokens deposited to DAppChain Gateway...`);
                while (pendingReceipt === null || pendingReceipt.oracleSignature.length === 0) {
                    pendingReceipt = yield this.getPendingWithdrawalReceiptAsync();
                    yield (0, helpers_1.sleep)(2000);
                }
            }
            signature = pendingReceipt.oracleSignature;
            return _1.CryptoUtils.bytesToHexAddr(signature);
        });
    }
    getUnclaimedLoomTokensAsync(owner) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const address = owner
                ? _1.Address.fromString(`eth:${owner}`)
                : _1.Address.fromString(`eth:${this.ethAddress}`);
            const tokens = yield this._dappchainGateway.getUnclaimedTokensAsync(address);
            const unclaimedLoomTokens = tokens.filter(t => t.tokenContract.local.toString() === this.ethereumLoom.address);
            // There is only 1 LOOM token and there's only 1 balance for it:
            // All other parameters of UnclaimedToken are for ERC721(x) tokens.
            let amount;
            if (unclaimedLoomTokens.length === 0) {
                // no unclaimed tokens
                amount = new bn_js_1.default(0);
            }
            else {
                // if the amounts array was set
                const amounts = unclaimedLoomTokens[0].tokenAmounts;
                amount = amounts ? amounts[0] : new bn_js_1.default(0);
            }
            return amount;
        });
    }
    withdrawCoinFromDAppChainGatewayAsync(amount, sig) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            if (this._version === 2 && this._ethereumVMC !== undefined) {
                const hash = yield this.createWithdrawalHash(amount);
                const validators = yield this._ethereumVMC.getValidators();
                const { vs, rs, ss, valIndexes } = (0, helpers_1.parseSigs)(sig, hash, validators);
                return this._ethereumGateway.functions.withdrawERC20(amount.toString(), this._ethereumLoom.address, valIndexes, vs, rs, ss);
            }
            // @ts-ignore
            return this._ethereumGateway.functions.withdrawERC20(amount.toString(), sig, this._ethereumLoom.address);
        });
    }
    createWithdrawalHash(amount) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            let nonce = yield this.ethereumGateway.functions.nonces(this.ethAddress);
            let amountHashed = ethers_1.ethers.utils.solidityKeccak256(['uint256', 'address'], [amount.toString(), this.ethereumLoom.address]);
            const msg = ethers_1.ethers.utils.solidityKeccak256(['string', 'address', 'uint256', 'address', 'bytes32'], [ERC20Prefix, this.ethAddress, nonce, this.ethereumGateway.address, amountHashed]);
            return msg;
        });
    }
    /**
     * Helper function to prefix an address with the chainId to get chainId:address format
     */
    prefixAddress(address) {
        return _1.Address.fromString(`${this.client.chainId}:${address}`);
    }
}
exports.GatewayUser = GatewayUser;
//# sourceMappingURL=gateway-user.js.map