"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.HTTPRPCClient = void 0;
const tslib_1 = require("tslib");
const axios_1 = tslib_1.__importDefault(require("axios"));
const events_1 = tslib_1.__importDefault(require("events"));
const debug_1 = tslib_1.__importDefault(require("debug"));
const log = (0, debug_1.default)('http-rpc-client');
/**
 * Sends JSON-RPC messages via HTTP.
 * Doesn't support listening to DAppChain events at the moment.
 */
class HTTPRPCClient extends events_1.default {
    get isSubscribed() {
        return false;
    }
    /**
     *
     * @param url
     * @param opts Options object
     * @param opts.requestTimeout Number of milliseconds to wait for a network operation to complete.
     */
    constructor(url, opts = {}) {
        super();
        this.url = url;
        this._rpcId = 0;
        this._getNextRequestId = () => (++this._rpcId).toString();
        const { requestTimeout = 15000, // 15s
        generateRequestId = this._getNextRequestId } = opts;
        this.requestTimeout = requestTimeout;
        // FIXME: generateRequestId doesn't actually override this._getNextRequestId at the moment
    }
    disconnect() {
        // no persistent connection, so do nothing
    }
    ensureConnectionAsync() {
        // no persistent connection, so do nothing
        return Promise.resolve();
    }
    /**
     * Sends a JSON-RPC message.
     * @param method RPC method name.
     * @param params Parameter object or array.
     * @returns A promise that will be resolved with the value of the result field (if any) in the
     *          JSON-RPC response message.
     */
    sendAsync(method, params) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            log(`Sending RPC msg to ${this.url}, method ${method}`);
            const req = {
                jsonrpc: '2.0',
                method,
                params,
                id: this._getNextRequestId()
            };
            const resp = yield axios_1.default.post(this.url, req, {
                timeout: this.requestTimeout
            });
            if (resp.data.error) {
                const { code, message, data } = resp.data.error;
                throw new Error(`JSON-RPC Error ${code} (${message}): ${data}`);
            }
            return resp.data.result;
        });
    }
}
exports.HTTPRPCClient = HTTPRPCClient;
//# sourceMappingURL=http-rpc-client.js.map