"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.WSRPCClient = void 0;
const tslib_1 = require("tslib");
const rpc_websockets_1 = require("rpc-websockets");
const events_1 = require("events");
const debug_1 = tslib_1.__importDefault(require("debug"));
const json_rpc_client_1 = require("./json-rpc-client");
const log = (0, debug_1.default)('ws-rpc-client');
/**
 * Sends JSON-RPC messages via web sockets.
 */
class WSRPCClient extends events_1.EventEmitter {
    get isSubscribed() {
        return this._isSubcribed;
    }
    /**
     *
     * @param url
     * @param opts Options object
     * @param opts.requestTimeout Number of milliseconds to wait for a network operation to complete.
     * @param opts.reconnectInterval Number of milliseconds to wait before attempting to reconnect
     *                               (in case the connection drops out).
     * @param opts.maxReconnects Maximum number of times to reconnect, defaults to infinity.
     */
    constructor(url, opts = {}) {
        super();
        this.url = url;
        this._isSubcribed = false;
        this._rpcId = 0;
        this._getNextRequestId = () => {
            const id = ++this._rpcId;
            return this.isWeb3EndpointEnabled ? id.toString() : id;
        };
        this._onEventMessage = (message) => {
            const msgStr = message instanceof ArrayBuffer ? Buffer.from(message).toString() : message;
            const msg = JSON.parse(msgStr);
            if (msg.method === 'eth_subscription') {
                log('EVM contract event arrived', msg);
                this.emit(json_rpc_client_1.RPCClientEvent.EVMMessage, this.url, msg);
            }
            else if (msg.id === '0') {
                // Events from native loomchain have id 0
                log('Go contract event arrived', msg);
                this.emit(json_rpc_client_1.RPCClientEvent.Message, this.url, msg);
            }
            else if (/^0x.+$/.test(msg.id)) {
                // Events from EVM have the id from the evmsubscribe command
                this.emit(json_rpc_client_1.RPCClientEvent.EVMMessage, this.url, msg);
            }
        };
        this.isWeb3EndpointEnabled = true === /eth$/.test(url);
        const { autoConnect = true, requestTimeout = 15000, // 15s
        reconnectInterval, maxReconnects = 0, // 0 means there is no limit
        generateRequestId = this._getNextRequestId } = opts;
        this._client = new rpc_websockets_1.Client(url, {
            autoconnect: autoConnect,
            reconnect: true,
            reconnect_interval: reconnectInterval,
            max_reconnects: maxReconnects
        }, generateRequestId);
        this.requestTimeout = requestTimeout;
        this.on('newListener', (event) => {
            if (event === json_rpc_client_1.RPCClientEvent.Message && this.listenerCount(event) === 0) {
                // rpc-websockets is just going to throw away the event messages from the DAppChain because
                // they don't conform to it's idea of notifications or events... fortunately few things in
                // javascript are truly private... so we'll just handle those event message ourselves ;)
                ;
                this._client.socket.on('message', this._onEventMessage);
                if (this._client.ready) {
                    log('Subscribe for events without topics');
                    this._client
                        .call('subevents', { topics: null }, this.requestTimeout)
                        .then(() => {
                        this._isSubcribed = true;
                        this.emit(json_rpc_client_1.RPCClientEvent.Subscribed, this.url, true);
                    })
                        .catch(err => this.emit(json_rpc_client_1.RPCClientEvent.Error, this.url, err));
                }
            }
            if (event === json_rpc_client_1.RPCClientEvent.EVMMessage && this.listenerCount(event) === 0) {
                ;
                this._client.socket.on('message', this._onEventMessage);
            }
        });
        this.on('removeListener', (event) => {
            if (event === json_rpc_client_1.RPCClientEvent.Message && this.listenerCount(event) === 0) {
                ;
                this._client.socket.removeListener('message', this._onEventMessage);
                if (this._client.ready) {
                    log('Unsubscribed for events');
                    this._client
                        .call('unsubevents', { topics: null }, this.requestTimeout)
                        .then(() => {
                        this._isSubcribed = false;
                        this.emit(json_rpc_client_1.RPCClientEvent.Subscribed, this.url, false);
                    })
                        .catch(err => {
                        this.emit(json_rpc_client_1.RPCClientEvent.Error, this.url, err);
                    });
                }
            }
            if (event === json_rpc_client_1.RPCClientEvent.EVMMessage && this.listenerCount(event) === 0) {
                ;
                this._client.socket.removeListener('message', this._onEventMessage);
            }
        });
        this._client.on('open', () => {
            this.emit(json_rpc_client_1.RPCClientEvent.Connected, this.url);
            if (this.listenerCount(json_rpc_client_1.RPCClientEvent.Message) > 0) {
                log('Subscribe for events without topics');
                this._client
                    .call('subevents', { topics: null }, this.requestTimeout)
                    .then(() => {
                    this._isSubcribed = true;
                    this.emit(json_rpc_client_1.RPCClientEvent.Subscribed, this.url, true);
                })
                    .catch(err => this.emit(json_rpc_client_1.RPCClientEvent.Error, this.url, err));
            }
        });
        this._client.on('close', () => {
            if (this.listenerCount(json_rpc_client_1.RPCClientEvent.Message) > 0) {
                this._isSubcribed = false;
                this.emit(json_rpc_client_1.RPCClientEvent.Subscribed, this.url, false);
            }
            this.emit(json_rpc_client_1.RPCClientEvent.Disconnected, this.url);
        });
        this._client.on('error', err => this.emit(json_rpc_client_1.RPCClientEvent.Error, this.url, err));
    }
    /**
     * Gracefully closes the underlying web socket connection.
     */
    disconnect() {
        // if `max_reconnects` is zero the client will attempt to reconnect even if `reconnect` is `false`,
        // this seems like a bug in rpc-websockets, but the workaround is simple enough...
        this._client.max_reconnects = 1;
        this._client.reconnect = false;
        this._client.close(0);
    }
    /**
     * Waits for a connection to be established to the server (if it isn't already).
     * @returns A promise that will be resolved when a connection is established.
     */
    ensureConnectionAsync() {
        if (this._client.ready) {
            return Promise.resolve();
        }
        return new Promise((resolve, reject) => {
            const timeout = setTimeout(() => reject(new Error('[WSRPCClient] Timeout while waiting for connection')), this.requestTimeout);
            this._client.once('open', () => {
                clearTimeout(timeout);
                resolve();
            });
        });
    }
    /**
     * Sends a JSON-RPC message.
     * @param method RPC method name.
     * @param params Parameter object or array.
     * @returns A promise that will be resolved with the value of the result field (if any) in the
     *          JSON-RPC response message.
     */
    sendAsync(method, params) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            yield this.ensureConnectionAsync();
            log(`Sending RPC msg to ${this.url}, method ${method}`);
            return this._client.call(method, params, this.requestTimeout);
        });
    }
}
exports.WSRPCClient = WSRPCClient;
//# sourceMappingURL=ws-rpc-client.js.map