"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.LoomProvider = void 0;
const tslib_1 = require("tslib");
const debug_1 = tslib_1.__importDefault(require("debug"));
const bn_js_1 = tslib_1.__importDefault(require("bn.js"));
const ethereumjs_util_1 = require("ethereumjs-util");
const retry_1 = tslib_1.__importDefault(require("retry"));
const client_1 = require("./client");
const helpers_1 = require("./helpers");
const loom_pb_1 = require("./proto/loom_pb");
const evm_pb_1 = require("./proto/evm_pb");
const address_1 = require("./address");
const crypto_utils_1 = require("./crypto-utils");
const solidity_helpers_1 = require("./solidity-helpers");
const big_uint_1 = require("./big-uint");
const log = (0, debug_1.default)('loom-provider');
const error = (0, debug_1.default)('loom-provider:error');
const bytesToHexAddrLC = (bytes) => {
    return (0, crypto_utils_1.bytesToHexAddr)(bytes).toLowerCase();
};
const numberToHexLC = (num) => {
    return (0, crypto_utils_1.numberToHex)(num).toLowerCase();
};
/**
 * Web3 provider that interacts with EVM contracts deployed on Loom DAppChains.
 */
class LoomProvider {
    /**
     * Constructs the LoomProvider to bridges communication between Web3 and Loom DappChains
     *
     * @param client Client from LoomJS
     * @param privateKey Account private key
     */
    constructor(client, privateKey, setupMiddlewaresFunction) {
        this._subscribed = false;
        /**
         * The retry strategy that should be used to retry some web3 requests.
         * By default failed requested won't be resent.
         * To understand how to tweak the retry strategy see
         * https://github.com/tim-kos/node-retry#retrytimeoutsoptions
         */
        this.retryStrategy = {
            retries: 0,
            minTimeout: 1000,
            maxTimeout: 30000,
            randomize: true
        };
        /**
         * Overrides the chain ID of the caller, when this is `null` the caller chain ID defaults
         * to the client chain ID.
         */
        this.callerChainId = null;
        this._client = client;
        this._netVersionFromChainId = LoomProvider.chainIdToNetVersion(this._client.chainId);
        this._setupMiddlewares = setupMiddlewaresFunction;
        this._ethSubscriptions = new Map();
        this._accountMiddlewares = new Map();
        this._ethRPCMethods = new Map();
        this.notificationCallbacks = new Array();
        this.accounts = new Map();
        // Only subscribe for event emitter do not call subevents
        this._client.addListener(client_1.ClientEvent.EVMEvent, (msg) => this._onWebSocketMessage(msg));
        if (!this._setupMiddlewares) {
            this._setupMiddlewares = (client, privateKey) => {
                return (0, helpers_1.createDefaultTxMiddleware)(client, privateKey);
            };
        }
        this._addDefaultMethods();
        if (client.isWeb3EndpointEnabled) {
            this._addWeb3EndpointMethods();
        }
        else {
            this._addLegacyDefaultMethods();
        }
        this._addDefaultEvents();
        this.addAccounts([privateKey]);
    }
    /**
     * Creates new accounts by passing the private key array
     *
     * Accounts will be available on public properties accounts
     *
     * @param accountsPrivateKey Array of private keys to create new accounts
     */
    addAccounts(accountsPrivateKey) {
        accountsPrivateKey.forEach(accountPrivateKey => {
            const publicKey = (0, crypto_utils_1.publicKeyFromPrivateKey)(accountPrivateKey);
            const accountAddress = address_1.LocalAddress.fromPublicKey(publicKey).toString();
            this.accounts.set(accountAddress, accountPrivateKey);
            this._accountMiddlewares.set(accountAddress, this._setupMiddlewares(this._client, accountPrivateKey));
            log(`New account added ${accountAddress}`);
        });
    }
    /**
     * Set an array of middlewares for a given account
     *
     * @param address Address to be register the middleware
     * @param middlewares Array of middlewares for the address
     */
    setMiddlewaresForAddress(address, middlewares) {
        this.accounts.set(address.toLowerCase(), null);
        this._accountMiddlewares.set(address.toLowerCase(), middlewares);
    }
    get accountMiddlewares() {
        return this._accountMiddlewares;
    }
    // PUBLIC FUNCTION TO SUPPORT WEB3
    on(type, callback) {
        log('on', type);
        switch (type) {
            case 'data':
                this.notificationCallbacks.push(callback);
                break;
            case 'connect':
                this._client.addListener(client_1.ClientEvent.Connected, callback);
                break;
            case 'end':
                this._client.addListener(client_1.ClientEvent.Disconnected, callback);
                break;
            case 'error':
                this._client.addListener(client_1.ClientEvent.Error, callback);
                break;
        }
    }
    _addDefaultEvents() {
        this._client.addListener(client_1.ClientEvent.Disconnected, () => {
            // reset all requests and callbacks
            this.reset();
        });
    }
    _addDefaultMethods() {
        this._ethRPCMethods.set('eth_accounts', this._ethAccounts);
        this._ethRPCMethods.set('eth_estimateGas', this._ethEstimateGas);
        this._ethRPCMethods.set('eth_gasPrice', this._ethGasPrice);
        this._ethRPCMethods.set('eth_newPendingTransactionFilter', this._ethNewPendingTransactionFilter);
        this._ethRPCMethods.set('eth_sendTransaction', this._ethSendTransaction);
        this._ethRPCMethods.set('eth_sign', this._ethSign);
        this._ethRPCMethods.set('personal_sign', this._personalSign);
        this._ethRPCMethods.set('net_version', this._netVersion);
    }
    /**
     * Sets up the provider to interact with the Loom /query endpoint, which requires Web3 JSON-RPC
     * messages to be marshalled to protobufs.
     */
    _addLegacyDefaultMethods() {
        this._ethRPCMethods.set('eth_blockNumber', this._ethBlockNumber);
        this._ethRPCMethods.set('eth_call', this._ethCall);
        this._ethRPCMethods.set('eth_getBalance', this._ethGetBalance);
        this._ethRPCMethods.set('eth_getBlockByHash', this._ethGetBlockByHash);
        this._ethRPCMethods.set('eth_getBlockByNumber', this._ethGetBlockByNumber);
        this._ethRPCMethods.set('eth_getCode', this._ethGetCode);
        this._ethRPCMethods.set('eth_getFilterChanges', this._ethGetFilterChanges);
        this._ethRPCMethods.set('eth_getLogs', this._ethGetLogs);
        this._ethRPCMethods.set('eth_getTransactionByHash', this._ethGetTransactionByHash);
        this._ethRPCMethods.set('eth_getTransactionReceipt', this._ethGetTransactionReceipt);
        this._ethRPCMethods.set('eth_newBlockFilter', this._ethNewBlockFilter);
        this._ethRPCMethods.set('eth_newFilter', this._ethNewFilter);
        this._ethRPCMethods.set('eth_uninstallFilter', this._ethUninstallFilter);
        this._ethRPCMethods.set('eth_subscribe', this._ethSubscribeLegacy);
        this._ethRPCMethods.set('eth_unsubscribe', this._ethUnsubscribeLegacy);
    }
    /**
     * Sets up the provider to interact with the Loom /eth endpoint.
     * This endpoint emulates the Web3 JSON-RPC API so most messages don't require transformation.
     */
    _addWeb3EndpointMethods() {
        this._ethRPCMethods.set('eth_blockNumber', this._ethCallSupportedMethod);
        this._ethRPCMethods.set('eth_call', this._ethCallSupportedMethod);
        this._ethRPCMethods.set('eth_getBalance', this._ethCallSupportedMethod);
        this._ethRPCMethods.set('eth_getBlockByHash', this._ethCallSupportedMethod);
        this._ethRPCMethods.set('eth_getBlockByNumber', this._ethCallSupportedMethod);
        this._ethRPCMethods.set('eth_getCode', this._ethCallSupportedMethod);
        this._ethRPCMethods.set('eth_getFilterChanges', this._ethCallSupportedMethod);
        this._ethRPCMethods.set('eth_getLogs', this._ethCallSupportedMethod);
        this._ethRPCMethods.set('eth_getTransactionByHash', this._ethCallSupportedMethod);
        this._ethRPCMethods.set('eth_getTransactionReceipt', this._ethCallSupportedMethod);
        this._ethRPCMethods.set('eth_newBlockFilter', this._ethCallSupportedMethod);
        this._ethRPCMethods.set('eth_newFilter', this._ethCallSupportedMethod);
        this._ethRPCMethods.set('eth_uninstallFilter', this._ethCallSupportedMethod);
        this._ethRPCMethods.set('eth_getStorageAt', this._ethCallSupportedMethod);
        this._ethRPCMethods.set('eth_subscribe', this._ethSubscribe);
        this._ethRPCMethods.set('eth_unsubscribe', this._ethUnsubscribe);
    }
    /**
     * Adds custom methods to the provider when a particular method isn't supported
     *
     * Throws if the added method already exists
     *
     * @param method name of the method to be added
     * @param customMethodFn function that will implement the method
     */
    addCustomMethod(method, customMethodFn) {
        if (this._ethRPCMethods.has(method)) {
            throw Error('Method already exists');
        }
        this._ethRPCMethods.set(method, customMethodFn);
    }
    /**
     * Overwrites existing method on the provider
     *
     * Throws if the overwritten method doesn't exists
     *
     * @param method name of the method to be overwritten
     * @param customMethodFn function that will implement the method
     */
    overwriteMethod(method, customMethodFn) {
        if (!this._ethRPCMethods.has(method)) {
            throw Error('Method to overwrite do not exists');
        }
        this._ethRPCMethods.set(method, customMethodFn);
    }
    /**
     * Return the numerical representation of the ChainId
     * More details at: https://github.com/loomnetwork/loom-js/issues/110
     */
    static chainIdToNetVersion(chainId) {
        // Avoids the error "Number can only safely store up to 53 bits" on Web3
        // Ensures the value less than 9007199254740991 (Number.MAX_SAFE_INTEGER)
        const chainIdHash = (0, solidity_helpers_1.soliditySha3)(chainId)
            .slice(2) // Removes 0x
            .slice(0, 13); // Produces safe Number less than 9007199254740991
        return new bn_js_1.default(chainIdHash).toNumber();
    }
    removeListener(type, callback) {
        switch (type) {
            case 'data':
                this.notificationCallbacks = [];
                break;
            case 'connect':
                this._client.removeListener(client_1.ClientEvent.Connected, callback);
                break;
            case 'end':
                this._client.removeListener(client_1.ClientEvent.Disconnected, callback);
                break;
            case 'error':
                this._client.removeListener(client_1.ClientEvent.Error, callback);
                break;
        }
    }
    removeAllListeners(type, callback) {
        if (type === 'data') {
            this.notificationCallbacks.forEach((cb, index) => {
                if (cb === callback) {
                    this.notificationCallbacks.splice(index, 1);
                }
            });
        }
    }
    disconnect() {
        this._client.disconnect();
    }
    /**
     * This function is invoked internally by Web3
     */
    reset() {
        this.notificationCallbacks = [];
    }
    // Adapter function for sendAsync from truffle provider
    sendAsync(payload, callback) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            if (callback) {
                return this.send(payload, callback);
            }
            else {
                return new Promise((resolve, reject) => {
                    this.send(payload, (err, result) => {
                        if (err)
                            reject(err);
                        else
                            resolve(result);
                    });
                });
            }
        });
    }
    /**
     * Should be used to make async request
     * This method is used internally by web3, so we adapt it to be used with loom contract
     * when we are wrapping the evm on a DAppChain
     * @param payload JSON payload generated by web3 which will be translated to loom transaction/call
     * @param callback Triggered on end with (err, result)
     */
    send(payload, callback) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            log('Request payload', JSON.stringify(payload, null, 2));
            const isArray = Array.isArray(payload);
            if (isArray) {
                payload = payload[0];
            }
            const prepareMethodToCall = (method) => {
                const methodToCall = this._ethRPCMethods.get(method);
                if (!methodToCall) {
                    throw Error(`Method "${payload.method}" not supported on this provider`);
                }
                return methodToCall;
            };
            try {
                // @ts-ignore
                const f = prepareMethodToCall(payload.method).bind(this);
                const result = yield f(payload);
                callback(null, this._okResponse(payload.id, result, isArray));
            }
            catch (err) {
                error(err);
                callback(err, null);
            }
        });
    }
    // PRIVATE FUNCTIONS EVM CALLS
    _ethCallSupportedMethod(payload) {
        return this._client.sendWeb3MsgAsync(payload.method, payload.params);
    }
    _ethAccounts() {
        if (this.accounts.size === 0) {
            throw Error('No account available');
        }
        const accounts = new Array();
        this.accounts.forEach((_value, key) => {
            accounts.push(key);
        });
        return accounts;
    }
    _ethBlockNumber() {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const blockNumber = yield this._client.getBlockHeightAsync();
            return numberToHexLC(+blockNumber);
        });
    }
    _ethCall(payload) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            // Sending a static call to Loom DAppChain
            const result = yield this._callStaticAsync(payload.params[0]);
            return result ? bytesToHexAddrLC(result) : '0x0';
        });
    }
    _ethEstimateGas() {
        // Loom DAppChain doesn't estimate gas
        // This method can be overwritten if necessary
        return null; // Returns null to afford with Web3 calls
    }
    _ethGetBalance() {
        // Loom DAppChain doesn't have ETH balance by default
        // This method can be overwritten if necessary
        return '0x0'; // Returns 0x0 to afford with Web3 calls
    }
    _ethGasPrice() {
        // Loom DAppChain doesn't use gas price
        // This method can be overwritten if necessary
        return null; // Returns null to afford with Web3 calls
    }
    _ethGetBlockByHash(payload) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const blockHash = payload.params[0];
            const isFull = payload.params[1] || true;
            const result = yield this._client.getEvmBlockByHashAsync(blockHash, isFull);
            if (!result) {
                return null;
            }
            return this._createBlockInfo(result, isFull);
        });
    }
    _ethGetBlockByNumber(payload) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const blockNumberToSearch = payload.params[0] === 'latest' ? payload.params[0] : (0, crypto_utils_1.hexToNumber)(payload.params[0]);
            const isFull = payload.params[1] || true;
            const result = yield this._client.getEvmBlockByNumberAsync(`${blockNumberToSearch}`, isFull);
            if (!result) {
                return null;
            }
            return this._createBlockInfo(result, isFull);
        });
    }
    _ethGetCode(payload) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const address = new address_1.Address(this._client.chainId, address_1.LocalAddress.fromHexString(payload.params[0]));
            const result = yield this._client.getEvmCodeAsync(address);
            if (!result) {
                throw Error('No code returned on eth_getCode');
            }
            return bytesToHexAddrLC(result);
        });
    }
    _ethGetFilterChanges(payload) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const result = yield this._client.getEvmFilterChangesAsync(payload.params[0]);
            if (!result) {
                return [];
            }
            if (result instanceof evm_pb_1.EthBlockHashList) {
                return result.getEthBlockHashList_asU8().map((hash) => bytesToHexAddrLC(hash));
            }
            else if (result instanceof evm_pb_1.EthTxHashList) {
                return result.getEthTxHashList_asU8().map((hash) => bytesToHexAddrLC(hash));
            }
            else if (result instanceof evm_pb_1.EthFilterLogList) {
                return result.getEthBlockLogsList().map((log) => this._createLogResult(log));
            }
        });
    }
    _ethGetLogs(payload) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            return this._getLogs(payload.params[0]);
        });
    }
    _ethGetTransactionByHash(payload) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            return this._getTransaction(payload.params[0]);
        });
    }
    _ethGetTransactionReceipt(payload) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const txHash = payload.params[0];
            const data = Buffer.from(txHash.slice(2), 'hex');
            const op = retry_1.default.operation(this.retryStrategy);
            const receipt = yield new Promise((resolve, reject) => {
                op.attempt(currentAttempt => {
                    log(`Current attempt ${currentAttempt}`);
                    this._client
                        .getEvmTxReceiptAsync((0, crypto_utils_1.bufferToProtobufBytes)(data))
                        .then(receipt => {
                        if (receipt) {
                            resolve(receipt);
                        }
                        else {
                            const err = new Error('Receipt cannot be empty');
                            error(err.message);
                            if (!op.retry(err)) {
                                reject(err);
                            }
                        }
                    })
                        .catch(err => {
                        if (!op.retry(err)) {
                            reject(err);
                        }
                        else {
                            error(err.message);
                        }
                    });
                });
            });
            return this._createReceiptResult(receipt);
        });
    }
    _ethNewBlockFilter() {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const result = yield this._client.newBlockEvmFilterAsync();
            if (!result) {
                throw Error('New block filter unexpected result');
            }
            return result;
        });
    }
    _ethNewFilter(payload) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const result = yield this._client.newEvmFilterAsync(payload.params[0]);
            if (!result) {
                throw Error('Cannot create new filter on eth_newFilter');
            }
            return result;
        });
    }
    _ethNewPendingTransactionFilter() {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const result = yield this._client.newPendingTransactionEvmFilterAsync();
            if (!result) {
                throw Error('New pending transaction filter unexpected result');
            }
            return result;
        });
    }
    _ethSendTransaction(payload) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            let result;
            if (payload.params[0].to) {
                result = yield this._callAsync(payload.params[0]);
            }
            else {
                result = yield this._deployAsync(payload.params[0]);
            }
            return bytesToHexAddrLC(result);
        });
    }
    _ethSign(payload) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const address = payload.params[0];
            const msg = payload.params[1];
            const privateKey = this.accounts.get(address);
            if (!privateKey) {
                throw Error('Account is not valid, private key not found');
            }
            const hash = (0, solidity_helpers_1.soliditySha3)('\x19Ethereum Signed Message:\n32', msg).slice(2);
            const privateHash = (0, solidity_helpers_1.soliditySha3)(privateKey).slice(2);
            const sig = (0, ethereumjs_util_1.ecsign)(Buffer.from(hash, 'hex'), Buffer.from(privateHash, 'hex'));
            return bytesToHexAddrLC(Buffer.concat([sig.r, sig.s, (0, ethereumjs_util_1.toBuffer)(sig.v)]));
        });
    }
    _personalSign(payload) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            // order of address & message params passed to personal_sign is inverse of eth_sign
            payload.params = [payload.params[1], payload.params[0]];
            yield this._ethSign(payload);
        });
    }
    _ethSubscribeLegacy(payload) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            if (!this._subscribed) {
                this._subscribed = true;
                this._client.addListenerForTopics().catch(console.error);
            }
            const method = payload.params[0];
            const filterObject = payload.params[1] || {};
            const result = yield this._client.evmSubscribeAsync(method, filterObject);
            if (!result) {
                throw Error('Subscribe filter failed');
            }
            return result;
        });
    }
    _ethSubscribe(payload) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            if (!this._subscribed) {
                this._subscribed = true;
                this._client.addListenerForTopics().catch(console.error);
            }
            const { method, params } = payload;
            const subscriptionId = yield this._client.evmSubscribeAsync(method, params);
            log('subscribed', subscriptionId, params);
            this._ethSubscriptions.set(subscriptionId, {
                id: subscriptionId,
                method,
                params
            });
            return subscriptionId;
        });
    }
    _ethUninstallFilter(payload) {
        return this._client.uninstallEvmFilterAsync(payload.params[0]);
    }
    _ethUnsubscribeLegacy(payload) {
        return this._client.evmUnsubscribeAsync(payload.params[0]);
    }
    _ethUnsubscribe(payload) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const subscriptionId = payload.params[0];
            const unsubscribeMethod = payload.params[1] || 'eth_unsubscribe';
            if (this._ethSubscriptions.has(subscriptionId)) {
                const response = yield this._client.sendWeb3MsgAsync(unsubscribeMethod, [subscriptionId]);
                if (response) {
                    this._ethSubscriptions.delete(subscriptionId);
                }
                return response;
            }
            throw new Error(`Provider error: Subscription with ID ${subscriptionId} does not exist.`);
        });
    }
    _netVersion() {
        return this._netVersionFromChainId;
    }
    // PRIVATE FUNCTIONS IMPLEMENTATIONS
    _deployAsync(payload) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const caller = new address_1.Address(this._client.chainId, address_1.LocalAddress.fromHexString(payload.from));
            const address = new address_1.Address(this._client.chainId, address_1.LocalAddress.fromHexString('0x0000000000000000000000000000000000000000'));
            const hasHexPrefix = payload.data.substring(0, 2) === '0x';
            const data = Buffer.from(payload.data.slice(hasHexPrefix ? 2 : 0), 'hex');
            const deployTx = new loom_pb_1.DeployTx();
            deployTx.setVmType(loom_pb_1.VMType.EVM);
            deployTx.setCode((0, crypto_utils_1.bufferToProtobufBytes)(data));
            const msgTx = new loom_pb_1.MessageTx();
            msgTx.setFrom(caller.MarshalPB());
            msgTx.setTo(address.MarshalPB());
            msgTx.setData(deployTx.serializeBinary());
            const tx = new loom_pb_1.Transaction();
            tx.setId(1);
            tx.setData(msgTx.serializeBinary());
            const ret = yield this._commitTransaction(payload.from, tx);
            const response = loom_pb_1.DeployResponse.deserializeBinary((0, crypto_utils_1.bufferToProtobufBytes)(ret));
            const responseData = loom_pb_1.DeployResponseData.deserializeBinary((0, crypto_utils_1.bufferToProtobufBytes)(response.getOutput_asU8()));
            return responseData.getTxHash_asU8();
        });
    }
    _callAsync(payload) {
        const chainId = this.callerChainId === null ? this._client.chainId : this.callerChainId;
        const caller = new address_1.Address(chainId, address_1.LocalAddress.fromHexString(payload.from));
        log('caller', caller.toString());
        const address = new address_1.Address(this._client.chainId, address_1.LocalAddress.fromHexString(payload.to));
        const data = Buffer.from(payload.data.slice(2), 'hex');
        const value = new bn_js_1.default((payload.value || '0x0').slice(2), 16);
        const callTx = new loom_pb_1.CallTx();
        callTx.setVmType(loom_pb_1.VMType.EVM);
        callTx.setInput((0, crypto_utils_1.bufferToProtobufBytes)(data));
        callTx.setValue((0, big_uint_1.marshalBigUIntPB)(value));
        const msgTx = new loom_pb_1.MessageTx();
        msgTx.setFrom(caller.MarshalPB());
        msgTx.setTo(address.MarshalPB());
        msgTx.setData(callTx.serializeBinary());
        const tx = new loom_pb_1.Transaction();
        tx.setId(2);
        tx.setData(msgTx.serializeBinary());
        return this._commitTransaction(payload.from, tx);
    }
    _callStaticAsync(payload) {
        const chainId = this.callerChainId === null ? this._client.chainId : this.callerChainId;
        const caller = new address_1.Address(chainId, address_1.LocalAddress.fromHexString(payload.from));
        log('caller', caller.toString());
        const address = new address_1.Address(this._client.chainId, address_1.LocalAddress.fromHexString(payload.to));
        const data = Buffer.from(payload.data.slice(2), 'hex');
        return this._client.queryAsync(address, data, loom_pb_1.VMType.EVM, caller);
    }
    _createBlockInfo(blockInfo, isFull) {
        const blockNumber = numberToHexLC(blockInfo.getNumber());
        const transactionHash = bytesToHexAddrLC(blockInfo.getHash_asU8());
        const parentHash = bytesToHexAddrLC(blockInfo.getParentHash_asU8());
        const logsBloom = bytesToHexAddrLC(blockInfo.getLogsBloom_asU8());
        const timestamp = blockInfo.getTimestamp();
        const transactions = blockInfo.getTransactionsList_asU8().map((transaction) => {
            if (isFull) {
                return this._createTransactionResult(evm_pb_1.EvmTxObject.deserializeBinary((0, crypto_utils_1.bufferToProtobufBytes)(transaction)));
            }
            else {
                return bytesToHexAddrLC(transaction);
            }
        });
        return {
            blockNumber,
            number: blockNumber,
            transactionHash,
            parentHash,
            logsBloom,
            timestamp,
            transactions,
            gasLimit: '0x0',
            gasUsed: '0x0',
            size: '0x0'
        };
    }
    _createTransactionResult(txObject) {
        const hash = bytesToHexAddrLC(txObject.getHash_asU8());
        const nonce = numberToHexLC(txObject.getNonce());
        const blockHash = bytesToHexAddrLC(txObject.getBlockHash_asU8());
        const blockNumber = numberToHexLC(txObject.getBlockNumber());
        const transactionIndex = numberToHexLC(txObject.getTransactionIndex());
        const from = bytesToHexAddrLC(txObject.getFrom_asU8());
        const to = bytesToHexAddrLC(txObject.getTo_asU8());
        const value = `${txObject.getValue()}`;
        const gas = numberToHexLC(txObject.getGas());
        const gasPrice = numberToHexLC(txObject.getGasPrice());
        const input = bytesToHexAddrLC(txObject.getInput_asU8());
        return {
            hash,
            nonce,
            blockHash,
            blockNumber,
            transactionIndex,
            from,
            to,
            value,
            gas,
            gasPrice,
            input
        };
    }
    _createReceiptResult(receipt) {
        const transactionHash = bytesToHexAddrLC(receipt.getTxHash_asU8());
        const transactionIndex = numberToHexLC(receipt.getTransactionIndex());
        const blockHash = bytesToHexAddrLC(receipt.getBlockHash_asU8());
        const blockNumber = numberToHexLC(receipt.getBlockNumber());
        const contractAddress = bytesToHexAddrLC(receipt.getContractAddress_asU8());
        const logs = receipt.getLogsList().map((logEvent, index) => {
            const logIndex = numberToHexLC(index);
            let data = bytesToHexAddrLC(logEvent.getEncodedBody_asU8());
            if (data === '0x') {
                data = '0x0';
            }
            return {
                logIndex,
                address: contractAddress,
                blockHash,
                blockNumber,
                blockTime: logEvent.getBlockTime(),
                transactionHash: bytesToHexAddrLC(logEvent.getTxHash_asU8()),
                transactionIndex,
                type: 'mined',
                data,
                topics: logEvent.getTopicsList().map((topic) => topic.toLowerCase())
            };
        });
        return {
            transactionHash,
            transactionIndex,
            blockHash,
            blockNumber,
            contractAddress,
            gasUsed: numberToHexLC(receipt.getGasUsed()),
            cumulativeGasUsed: numberToHexLC(receipt.getCumulativeGasUsed()),
            logs,
            status: numberToHexLC(receipt.getStatus())
        };
    }
    _getTransaction(txHash) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const data = Buffer.from(txHash.slice(2), 'hex');
            const transaction = yield this._client.getEvmTxByHashAsync((0, crypto_utils_1.bufferToProtobufBytes)(data));
            if (!transaction) {
                throw Error('Transaction cannot be empty');
            }
            const hash = bytesToHexAddrLC(transaction.getHash_asU8());
            const nonce = numberToHexLC(transaction.getNonce());
            const transactionIndex = numberToHexLC(transaction.getTransactionIndex());
            const blockHash = bytesToHexAddrLC(transaction.getBlockHash_asU8());
            const blockNumber = numberToHexLC(transaction.getBlockNumber());
            const from = bytesToHexAddrLC(transaction.getFrom_asU8());
            const to = bytesToHexAddrLC(transaction.getTo_asU8());
            const value = numberToHexLC(transaction.getValue());
            const gasPrice = numberToHexLC(transaction.getGasPrice());
            const gas = numberToHexLC(transaction.getGas());
            const input = '0x0';
            return {
                hash,
                nonce,
                blockHash,
                blockNumber,
                transactionIndex,
                from,
                to,
                value,
                gasPrice,
                gas,
                input
            };
        });
    }
    _createLogResult(log) {
        return {
            removed: log.getRemoved(),
            blockTime: log.getBlockTime(),
            logIndex: numberToHexLC(log.getLogIndex()),
            transactionIndex: (0, crypto_utils_1.numberToHex)(log.getTransactionIndex()),
            transactionHash: bytesToHexAddrLC(log.getTransactionHash_asU8()),
            blockHash: bytesToHexAddrLC(log.getBlockHash_asU8()),
            blockNumber: (0, crypto_utils_1.numberToHex)(log.getBlockNumber()),
            address: bytesToHexAddrLC(log.getAddress_asU8()),
            data: bytesToHexAddrLC(log.getData_asU8()),
            topics: log.getTopicsList().map((topic) => String.fromCharCode.apply(null, topic))
        };
    }
    _getLogs(filterObject) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const logsListAsyncResult = yield this._client.getEvmLogsAsync(filterObject);
            if (!logsListAsyncResult) {
                return [];
            }
            const logList = evm_pb_1.EthFilterLogList.deserializeBinary((0, crypto_utils_1.bufferToProtobufBytes)(logsListAsyncResult));
            return logList.getEthBlockLogsList().map((log) => {
                return this._createLogResult(log);
            });
        });
    }
    _onWebSocketMessage(msgEvent) {
        if (this._client.isWeb3EndpointEnabled) {
            log('Socket message arrived (web3)', JSON.stringify(msgEvent, null, 2));
            this.notificationCallbacks.forEach(callback => callback(msgEvent));
            return;
        }
        const eventArgs = msgEvent;
        if (eventArgs.kind === client_1.ClientEvent.EVMEvent) {
            const JSONRPCResult = {
                jsonrpc: '2.0',
                method: 'eth_subscription',
                params: {
                    subscription: msgEvent.id,
                    result: {
                        transactionHash: bytesToHexAddrLC(eventArgs.transactionHashBytes),
                        logIndex: '0x0',
                        transactionIndex: '0x0',
                        blockHash: '0x0',
                        blockNumber: numberToHexLC(+eventArgs.blockHeight),
                        address: eventArgs.contractAddress.local.toString(),
                        type: 'mined',
                        data: bytesToHexAddrLC(eventArgs.data),
                        topics: eventArgs.topics
                    }
                }
            };
            log('Socket message arrived', JSON.stringify(JSONRPCResult, null, 2));
            this.notificationCallbacks.forEach(callback => callback(JSONRPCResult));
        }
    }
    _commitTransaction(fromPublicAddr, txTransaction) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const middleware = this._accountMiddlewares.get(fromPublicAddr);
            return this._client.commitTxAsync(txTransaction, { middleware });
        });
    }
    // Basic response to web3js
    _okResponse(id, result = 0, isArray = false) {
        const response = { id, jsonrpc: '2.0', result };
        const ret = isArray ? [response] : response;
        log('Response payload', JSON.stringify(ret, null, 2));
        return ret;
    }
}
exports.LoomProvider = LoomProvider;
//# sourceMappingURL=loom-provider.js.map