"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.CachedNonceTxMiddleware = void 0;
const tslib_1 = require("tslib");
const debug_1 = tslib_1.__importDefault(require("debug"));
const loom_pb_1 = require("../proto/loom_pb");
const client_1 = require("../client");
const crypto_utils_1 = require("../crypto-utils");
const nonce_tx_middleware_1 = require("./nonce-tx-middleware");
const address_1 = require("../address");
const log = (0, debug_1.default)('cached-nonce-tx-middleware');
/**
 * Wraps data in a NonceTx.
 * This middleware obtains the initial nonce value from the chain, and then increments it locally
 * for every tx, if a tx fails due to a nonce mismatch the chain is queried again to obtain the
 * latest nonce.
 */
class CachedNonceTxMiddleware {
    constructor(publicKeyOrAccount, client) {
        this._publicKey = null;
        this._account = null;
        if (publicKeyOrAccount instanceof address_1.Address) {
            this._account = publicKeyOrAccount;
        }
        else {
            this._publicKey = publicKeyOrAccount;
        }
        this._client = client;
        this._lastNonce = -1;
    }
    Handle(txData) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            if (this._lastNonce === -1) {
                log('Nonce not cached');
                try {
                    const key = this._publicKey ? (0, crypto_utils_1.bytesToHex)(this._publicKey) : undefined;
                    const account = this._account ? this._account.toString() : undefined;
                    this._lastNonce = yield this._client.getAccountNonceAsync({ key, account });
                }
                catch (err) {
                    throw Error('Failed to obtain latest nonce');
                }
            }
            log(`Next nonce ${this._lastNonce + 1}`);
            const tx = new loom_pb_1.NonceTx();
            tx.setInner(txData);
            tx.setSequence(this._lastNonce + 1);
            return tx.serializeBinary();
        });
    }
    HandleResults(results) {
        const { validation, commit } = results;
        const isInvalidTx = validation && validation.code !== 0;
        const isFailedTx = commit && commit.code;
        if (isInvalidTx || isFailedTx) {
            // Nonce has to be reset regardless of the cause of the tx failure.
            log(`Reset cached nonce due to failed tx`);
            this._lastNonce = -1;
            // Throw a specific error for a nonce mismatch
            const isCheckTxNonceInvalid = validation &&
                validation.code === 1 &&
                (validation.log && validation.log.indexOf('sequence number does not match') !== -1);
            const isDeliverTxNonceInvalid = commit &&
                commit.code === 1 &&
                (commit.log && commit.log.indexOf('sequence number does not match') !== -1);
            if (isCheckTxNonceInvalid || isDeliverTxNonceInvalid) {
                throw new Error(nonce_tx_middleware_1.INVALID_TX_NONCE_ERROR);
            }
        }
        else if (this._lastNonce !== -1) {
            // Only increment the nonce if the tx is valid
            this._lastNonce++;
            log(`Incremented cached nonce to ${this._lastNonce}`);
        }
        return results;
    }
    handleError(err) {
        if ((0, client_1.isTxAlreadyInCacheError)(err)) {
            // This error indicates the tx payload & nonce were identical to a previously sent tx,
            // which means the cached nonce has diverged from the nonce on the node, need to clear it out
            // so it's refetched for the next tx.
            this._lastNonce = -1;
            log('Reset cached nonce due to dupe tx');
        }
    }
}
exports.CachedNonceTxMiddleware = CachedNonceTxMiddleware;
//# sourceMappingURL=cached-nonce-tx-middleware.js.map