"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.NonceTxMiddleware = exports.isInvalidTxNonceError = exports.INVALID_TX_NONCE_ERROR = void 0;
const tslib_1 = require("tslib");
const debug_1 = tslib_1.__importDefault(require("debug"));
const address_1 = require("../address");
const crypto_utils_1 = require("../crypto-utils");
const loom_pb_1 = require("../proto/loom_pb");
const log = (0, debug_1.default)('nonce-tx-middleware');
exports.INVALID_TX_NONCE_ERROR = 'Invalid tx nonce';
function isInvalidTxNonceError(err) {
    return err instanceof Error && err.message === exports.INVALID_TX_NONCE_ERROR;
}
exports.isInvalidTxNonceError = isInvalidTxNonceError;
/**
 * Wraps data in a NonceTx.
 * This middleware obtains the latest nonce from the chain for each tx.
 * The Loom DAppChain keeps track of the nonce of the last committed tx to prevent replay attacks.
 */
class NonceTxMiddleware {
    constructor(publicKeyOrAccount, client) {
        this._publicKey = null;
        this._account = null;
        if (publicKeyOrAccount instanceof address_1.Address) {
            this._account = publicKeyOrAccount;
        }
        else {
            this._publicKey = publicKeyOrAccount;
        }
        this._client = client;
    }
    Handle(txData) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const key = this._publicKey ? (0, crypto_utils_1.bytesToHex)(this._publicKey) : undefined;
            const account = this._account ? this._account.toString() : undefined;
            const nonce = yield this._client.getAccountNonceAsync({ key, account });
            log(`Next nonce ${nonce + 1}`);
            const tx = new loom_pb_1.NonceTx();
            tx.setInner(txData);
            tx.setSequence(nonce + 1);
            return tx.serializeBinary();
        });
    }
    HandleResults(results) {
        const { validation, commit } = results;
        if (validation &&
            validation.code === 1 &&
            (validation.log && validation.log.indexOf('sequence number does not match') !== -1)) {
            throw new Error(exports.INVALID_TX_NONCE_ERROR);
        }
        if (commit &&
            commit.code === 1 &&
            (commit.log && commit.log.indexOf('sequence number does not match') !== -1)) {
            throw new Error(exports.INVALID_TX_NONCE_ERROR);
        }
        return results;
    }
}
exports.NonceTxMiddleware = NonceTxMiddleware;
//# sourceMappingURL=nonce-tx-middleware.js.map