"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.SignedEthTxMiddleware = void 0;
const tslib_1 = require("tslib");
const debug_1 = tslib_1.__importDefault(require("debug"));
const loom_pb_1 = require("../proto/loom_pb");
const solidity_helpers_1 = require("../solidity-helpers");
const crypto_utils_1 = require("../crypto-utils");
const log = (0, debug_1.default)('signed-eth-tx-middleware');
/**
 * Signs transactions using an Ethereum compatible (secp256k1) private key.
 */
class SignedEthTxMiddleware {
    /**
     * @param signer ethers.js signer to use for signing txs.
     */
    constructor(signer) {
        this.signer = signer;
    }
    Handle(txData) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            if (!this.signerAddress) {
                // Getting the public key address
                this.signerAddress = yield this.signer.getAddress();
            }
            // Get hash to be signed
            const hash = (0, solidity_helpers_1.soliditySha3)({
                type: 'bytes',
                value: (0, crypto_utils_1.bytesToHex)(txData)
            });
            // Signing the transaction
            const etherSigner = new solidity_helpers_1.EthersSigner(this.signer);
            const sig = yield etherSigner.signAsync(hash);
            log(`signer: ${this.signerAddress}, signature: 0x${(0, crypto_utils_1.bytesToHex)(sig.slice(1))}`);
            const signedTx = new loom_pb_1.SignedTx();
            signedTx.setInner(txData);
            signedTx.setSignature(sig);
            // NOTE: don't need to send secp256k1 public key sig since it can be recovered from hash + sig
            return signedTx.serializeBinary();
        });
    }
}
exports.SignedEthTxMiddleware = SignedEthTxMiddleware;
//# sourceMappingURL=signed-eth-tx-middleware.js.map