"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.selectProtocol = exports.createJSONRPCClient = void 0;
const tslib_1 = require("tslib");
const json_rpc_client_1 = require("./internal/json-rpc-client");
const ws_rpc_client_1 = require("./internal/ws-rpc-client");
const http_rpc_client_1 = require("./internal/http-rpc-client");
const dual_rpc_client_1 = require("./internal/dual-rpc-client");
/**
 * Creates an RPC client for communicating with a Loom DAppChain based on the specified options.
 * @param opts Options object
 * @param opts.protocols
 * @param opts.autoConnect If `true` the client will automatically connect after being created,
 *                         defaults to `true` and shouldn't be changed.
 * @param opts.requestTimeout Maximum number of milliseconds the client should wait for a request
 *                            to receive a response.
 * @param opts.generateRequestId Can be set to override the default JSON-RPC message ID generator.
 */
function createJSONRPCClient(opts) {
    const { protocols, autoConnect = true, requestTimeout, generateRequestId } = opts;
    if (protocols.length === 1) {
        const _a = protocols[0], { url } = _a, otherOpts = tslib_1.__rest(_a, ["url"]);
        const protocol = selectProtocol(url);
        if (protocol === json_rpc_client_1.JSONRPCProtocol.HTTP) {
            return new http_rpc_client_1.HTTPRPCClient(url, { requestTimeout, generateRequestId });
        }
        else if (protocol === json_rpc_client_1.JSONRPCProtocol.WS) {
            return new ws_rpc_client_1.WSRPCClient(url, Object.assign({ autoConnect, requestTimeout, generateRequestId }, otherOpts));
        }
    }
    else if (protocols.length === 2) {
        const p1 = selectProtocol(protocols[0].url);
        const p2 = selectProtocol(protocols[1].url);
        if (p1 === json_rpc_client_1.JSONRPCProtocol.HTTP && p2 === json_rpc_client_1.JSONRPCProtocol.WS) {
            const { reconnectInterval, maxReconnects } = protocols[1];
            return new dual_rpc_client_1.DualRPCClient({
                httpUrl: protocols[0].url,
                wsUrl: protocols[1].url,
                autoConnect,
                protocol: p1,
                requestTimeout,
                generateRequestId,
                reconnectInterval,
                maxReconnects
            });
        }
        else if (p2 === json_rpc_client_1.JSONRPCProtocol.HTTP && p1 === json_rpc_client_1.JSONRPCProtocol.WS) {
            const { reconnectInterval, maxReconnects } = protocols[0];
            return new dual_rpc_client_1.DualRPCClient({
                httpUrl: protocols[1].url,
                wsUrl: protocols[0].url,
                autoConnect,
                protocol: p1,
                requestTimeout,
                generateRequestId,
                reconnectInterval,
                maxReconnects
            });
        }
    }
    throw new Error('Failed to create JSON-RPC client: invalid protocol configuration');
}
exports.createJSONRPCClient = createJSONRPCClient;
function selectProtocol(url) {
    if (url.startsWith('http://') || url.startsWith('https://')) {
        return json_rpc_client_1.JSONRPCProtocol.HTTP;
    }
    else if (url.startsWith('ws://') || url.startsWith('wss://')) {
        return json_rpc_client_1.JSONRPCProtocol.WS;
    }
    else {
        throw new Error('Invalid URL');
    }
}
exports.selectProtocol = selectProtocol;
//# sourceMappingURL=rpc-client-factory.js.map