"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.OfflineWeb3Signer = exports.Web3Signer = exports.EthersSigner = exports.getJsonRPCSignerAsync = exports.getMetamaskSigner = exports.soliditySha3 = void 0;
const tslib_1 = require("tslib");
const ethereumjs_util_1 = tslib_1.__importDefault(require("ethereumjs-util"));
const web3_1 = tslib_1.__importDefault(require("web3"));
const ethers_1 = require("ethers");
const web3 = new web3_1.default();
function soliditySha3(...values) {
    return web3.utils.soliditySha3(...values);
}
exports.soliditySha3 = soliditySha3;
/**
 * Returns the Metamask signer from web3 current provider
 */
function getMetamaskSigner(provider) {
    // HACK: force personal sign by pretending to be metamask no matter what the web3 provider is
    provider.isMetaMask = true;
    return new ethers_1.ethers.providers.Web3Provider(provider).getSigner();
}
exports.getMetamaskSigner = getMetamaskSigner;
/**
 * Returns json rpc signer, ex: http://localhost:8545
 *
 * @param urlString url string to connect to provider
 * @param accountIndex index of the account on providers list
 */
function getJsonRPCSignerAsync(urlString, accountIndex = 0) {
    return tslib_1.__awaiter(this, void 0, void 0, function* () {
        const provider = new ethers_1.ethers.providers.JsonRpcProvider(urlString);
        const signers = (yield provider.listAccounts()).map((acc) => provider.getSigner(acc));
        return signers[accountIndex];
    });
}
exports.getJsonRPCSignerAsync = getJsonRPCSignerAsync;
/**
 * Signs message using a Web3 account.
 * This signer should be used for interactive signing in the browser with MetaMask.
 */
class EthersSigner {
    /**
     * @param web3 Web3 instance to use for signing.
     * @param accountAddress Address of web3 account to sign with.
     */
    constructor(signer) {
        this._signer = signer;
    }
    /**
     * Signs a message.
     * @param msg Message to sign.
     * @returns Promise that will be resolved with the signature bytes.
     */
    signAsync(msg) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            let flatSig = yield this._signer.signMessage(ethers_1.ethers.utils.arrayify(msg));
            const sig = ethers_1.ethers.utils.splitSignature(flatSig);
            let v = sig.v;
            if (v === 0 || v === 1) {
                v += 27;
            }
            flatSig = '0x01' + sig.r.slice(2) + sig.s.slice(2) + v.toString(16);
            return ethereumjs_util_1.default.toBuffer(flatSig);
        });
    }
}
exports.EthersSigner = EthersSigner;
/**
 * Signs message using a Web3 account.
 * This signer should be used for interactive signing in the browser with MetaMask.
 */
class Web3Signer {
    /**
     * @param web3 Web3 instance to use for signing.
     * @param accountAddress Address of web3 account to sign with.
     */
    constructor(web3, accountAddress) {
        this._web3 = web3;
        this._address = accountAddress;
    }
    /**
     * Signs a message.
     * @param msg Message to sign.
     * @returns Promise that will be resolved with the signature bytes.
     */
    signAsync(msg) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const signature = yield this._web3.eth.sign(msg, this._address);
            const sig = signature.slice(2);
            let mode = 1 /* SIGNATURE_TYPE.GETH */; // Geth
            const r = ethereumjs_util_1.default.toBuffer('0x' + sig.substring(0, 64));
            const s = ethereumjs_util_1.default.toBuffer('0x' + sig.substring(64, 128));
            let v = parseInt(sig.substring(128, 130), 16);
            if (v === 0 || v === 1) {
                v += 27;
            }
            else {
                mode = 0 /* SIGNATURE_TYPE.EIP712 */; // indicate that msg wasn't prefixed before signing (MetaMask doesn't prefix!)
            }
            return Buffer.concat([ethereumjs_util_1.default.toBuffer(mode), r, s, ethereumjs_util_1.default.toBuffer(v)]);
        });
    }
}
exports.Web3Signer = Web3Signer;
/**
 * THIS IS BEING DEPRECATED
 * Signs message using a Web3 account.
 * This signer should be used for signing in NodeJS.
 */
class OfflineWeb3Signer {
    /**
     * @param web3 Web3 instance to use for signing.
     * @param account Web3 account to sign with.
     */
    constructor(web3, account) {
        this._web3 = web3;
        this._account = account;
    }
    /**
     * Signs a message.
     * @param msg Message to sign.
     * @returns Promise that will be resolved with the signature bytes.
     */
    signAsync(msg) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const ret = yield this._web3.eth.accounts.sign(msg, this._account.privateKey);
            // @ts-ignore
            const sig = ret.signature.slice(2);
            const r = ethereumjs_util_1.default.toBuffer('0x' + sig.substring(0, 64));
            const s = ethereumjs_util_1.default.toBuffer('0x' + sig.substring(64, 128));
            let v = parseInt(sig.substring(128, 130), 16);
            return Buffer.concat([
                ethereumjs_util_1.default.toBuffer(1 /* SIGNATURE_TYPE.GETH */),
                r,
                s,
                ethereumjs_util_1.default.toBuffer(v)
            ]);
        });
    }
}
exports.OfflineWeb3Signer = OfflineWeb3Signer;
//# sourceMappingURL=solidity-helpers.js.map